from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Callable,
    Generator,
    Iterator,
    Optional,
    Set,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
import polars as pl
from tqdm import tqdm

from eir.data_load.label_setup import get_file_path_iterator
from eir.setup.input_setup_modules.setup_sequence import get_sequence_split_function
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    pass

logger = get_logger(name=__name__)


def get_file_sample_id_iterator_basic(
    data_source: str,
    ids_to_keep: Union[None, Set[str]],
) -> Generator[Tuple[str, Path], None, None]:
    base_file_iterator = get_file_path_iterator(
        data_source=Path(data_source), validate=False
    )

    for file in base_file_iterator:
        sample_id = file.stem

        if ids_to_keep:
            if sample_id in ids_to_keep:
                yield sample_id, file
        else:
            yield sample_id, file


def add_sequence_data_from_csv_to_df(
    input_source: str,
    input_df: pl.DataFrame,
    ids_to_keep: Union[None, Set[str]],
    split_on: Optional[str],
    encode_func: Callable,
    input_name: str = "CSV File Data",
) -> pl.DataFrame:
    logger.info(
        "Loading sequence data from CSV file %s. Note that this will "
        "load all the sequence data into memory.",
        input_source,
    )

    split_func = get_sequence_split_function(split_on=split_on)
    csv_sequence_iterator = get_csv_id_sequence_iterator(
        data_source=input_source,
        ids_to_keep=ids_to_keep,
    )

    ids = []
    sequences = []

    for sample_id, sequence in tqdm(csv_sequence_iterator, desc=input_name):
        sequence_split = split_func(sequence)
        sequence_encoded = encode_func(sequence_split)

        if isinstance(sequence_encoded, np.ndarray):
            sequence_encoded = sequence_encoded.tolist()

        ids.append(sample_id)
        sequences.append(sequence_encoded)

    if not ids:
        return input_df

    sequence_df = pl.DataFrame(
        {
            "ID": pl.Series(name="ID", values=ids, dtype=pl.Utf8),
            input_name: pl.Series(
                name=input_name, values=sequences, dtype=pl.List(pl.Int64)
            ),
        }
    )

    if input_df.height == 0:
        return sequence_df
    else:
        return input_df.join(sequence_df, on="ID", how="full", coalesce=True)


def get_csv_id_sequence_iterator(
    data_source: str, ids_to_keep: Optional[Set[str]] = None
) -> Iterator[Tuple[str, str]]:
    df = pd.read_csv(data_source, index_col="ID", dtype={"ID": str})

    if "Sequence" not in df.columns:
        raise ValueError(
            f"Expected to find a 'Sequence' column in {data_source}, but didn't."
        )

    if ids_to_keep:
        df = df[df.index.isin(set(ids_to_keep))]

    for row in df.itertuples():
        cur_seq = row.Sequence
        if pd.isna(cur_seq):
            cur_seq = ""

        cur_index = row.Index
        assert isinstance(cur_index, str)
        assert isinstance(cur_seq, str)

        yield cur_index, cur_seq



import argparse

import uvicorn
from fastapi import FastAPI

from eir.serve_modules.serve_api import create_info_endpoint, create_predict_endpoint
from eir.serve_modules.serve_experiment_io import (
    ServeExperiment,
    load_experiment_for_serve,
)
from eir.utils.logging import get_logger

logger = get_logger(name=__name__, tqdm_compatible=True)

app = FastAPI()


def load_experiment(model_path: str, device: str) -> ServeExperiment:
    serve_experiment = load_experiment_for_serve(
        model_path=model_path,
        device=device,
    )
    return serve_experiment


def get_cl_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Load a model for serving.")
    parser.add_argument(
        "--model-path",
        type=str,
        required=True,
        help="Path to the model file",
    )

    parser.add_argument(
        "--device",
        type=str,
        default="cpu",
        help="Device to run the model on",
    )

    parser.add_argument(
        "--port",
        type=int,
        default=8000,
        help="Port to run the server on",
    )

    parser.add_argument(
        "--host",
        type=str,
        default="0.0.0.0",
        help="Host to run the server on",
    )

    args = parser.parse_args()

    return args


def main():
    args = get_cl_args()

    serve_experiment = load_experiment(
        model_path=args.model_path,
        device=args.device,
    )
    logger.info(f"Model loaded from {args.model_path}")

    create_info_endpoint(
        app=app,
        serve_experiment=serve_experiment,
    )

    create_predict_endpoint(
        app=app,
        configs=serve_experiment.configs.input_configs,
        serve_experiment=serve_experiment,
    )

    uvicorn.run(app=app, host=args.host, port=args.port)


if __name__ == "__main__":
    main()



# protocol.py
from typing import Dict, List, TypedDict, Union


class GetInfoPayload(TypedDict):
    pass


class GetDataPayload(TypedDict):
    batch_size: int


class StatusPayload(TypedDict):
    pass


class ResetPayload(TypedDict):
    pass


class InfoPayload(TypedDict):
    dataset_info: Dict[str, Union[str, List[int]]]


class DataPayload(TypedDict):
    samples: List[Dict[str, Union[List[float], str]]]


class ResetConfirmationPayload(TypedDict):
    message: str


class ErrorPayload(TypedDict):
    message: str


class HeartbeatPayload(TypedDict):
    pass


class HandshakePayload(TypedDict):
    pass


class ValidationIdsConfirmationPayload(TypedDict):
    message: str


MESSAGE_TYPES = {
    "getInfo": {"direction": "client_to_server", "payload": GetInfoPayload},
    "getData": {"direction": "client_to_server", "payload": GetDataPayload},
    "status": {"direction": "bidirectional", "payload": StatusPayload},
    "reset": {"direction": "client_to_server", "payload": ResetPayload},
    "resetConfirmation": {
        "direction": "server_to_client",
        "payload": ResetConfirmationPayload,
    },
    "info": {"direction": "server_to_client", "payload": InfoPayload},
    "data": {"direction": "server_to_client", "payload": DataPayload},
    "error": {"direction": "bidirectional", "payload": ErrorPayload},
    "heartbeat": {"direction": "bidirectional", "payload": HeartbeatPayload},
    "handshake": {"direction": "bidirectional", "payload": HandshakePayload},
    "validationIdsConfirmation": {
        "direction": "server_to_client",
        "payload": ValidationIdsConfirmationPayload,
    },
}

FROM_SERVER_MESSAGE_TYPES = {
    "info",
    "data",
    "status",
    "heartbeat",
    "error",
    "handshake",
    "validationIdsConfirmation",
    "resetConfirmation",
}

PROTOCOL_VERSION = "1.0"



import json
from argparse import Namespace
from copy import copy
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Iterable, Literal, Sequence, Union

import numpy as np
import polars as pl
from aislib.misc_utils import ensure_path_exists
from torch.utils.data import DataLoader

from eir import train
from eir.data_load import datasets, label_setup
from eir.experiment_io.experiment_io import (
    LoadedTrainExperiment,
    load_serialized_train_experiment,
)
from eir.experiment_io.io_utils import check_version, get_run_folder_from_model_path
from eir.models.model_setup_modules.meta_setup import (
    al_meta_model,
    get_default_meta_class,
    get_meta_model_class_and_kwargs_from_configs,
)
from eir.models.model_setup_modules.model_io import load_model
from eir.models.model_training_utils import get_prediction_outputs_generator
from eir.predict_modules.predict_attributions import compute_predict_attributions
from eir.predict_modules.predict_config import converge_train_and_predict_configs
from eir.predict_modules.predict_data import set_up_default_dataset
from eir.predict_modules.predict_input_setup import set_up_inputs_for_predict
from eir.predict_modules.predict_output_modules.predict_array_output import (
    predict_array_wrapper,
)
from eir.predict_modules.predict_output_modules.predict_sequence_output import (
    predict_sequence_wrapper,
)
from eir.predict_modules.predict_output_modules.predict_survival_output import (
    predict_survival_wrapper_no_labels,
    predict_survival_wrapper_with_labels,
)
from eir.predict_modules.predict_output_modules.predict_tabular_output import (
    predict_tabular_wrapper_no_labels,
    predict_tabular_wrapper_with_labels,
)
from eir.predict_modules.predict_target_setup import (
    MissingTargetsInfo,
    get_target_labels_for_testing,
)
from eir.setup.config import Configs, get_main_parser
from eir.setup.config_setup_modules.config_setup_utils import load_yaml_config
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.output_setup import al_output_objects_as_dict
from eir.setup.schemas import OutputConfig
from eir.target_setup.target_label_setup import gather_all_ids_from_output_configs
from eir.train import check_dataset_and_batch_size_compatibility
from eir.train_utils.evaluation import (
    deregister_pre_evaluation_hooks,
    register_pre_evaluation_hooks,
    run_all_eval_hook_analysis,
    run_split_evaluation,
)
from eir.train_utils.metrics import al_metric_record_dict, al_step_metric_dict
from eir.train_utils.step_logic import Hooks, prepare_base_batch_default
from eir.train_utils.utils import set_log_level_for_eir_loggers
from eir.utils.logging import get_logger

logger = get_logger(name=__name__, tqdm_compatible=True)

al_predict_hook = Callable[..., dict]
al_predict_hooks = Iterable[al_predict_hook]


@dataclass()
class PredictSpecificCLArgs:
    model_path: str
    evaluate: bool
    output_folder: str
    attribution_background_source: Union[Literal["train"], Literal["predict"]]


def main():
    main_parser = get_main_parser(output_nargs="*", global_nargs="*")

    main_parser.add_argument(
        "--model_path",
        type=str,
        required=True,
        help="Path to model to use for predictions.",
    )

    main_parser.add_argument("--evaluate", dest="evaluate", action="store_true")

    main_parser.add_argument(
        "--output_folder",
        type=str,
        help="Where to save prediction results.",
        required=True,
    )

    main_parser.add_argument(
        "--attribution_background_source",
        type=str,
        help="For attribution analysis, whether to load backgrounds from the data used "
        "for training or to use the current data passed to the predict module.",
        choices=["train", "predict"],
        default="train",
    )

    predict_cl_args = main_parser.parse_args()

    _verify_predict_cl_args(predict_cl_args=predict_cl_args)
    run_folder = get_run_folder_from_model_path(model_path=predict_cl_args.model_path)
    check_version(run_folder=run_folder)

    run_predict(predict_cl_args=predict_cl_args)


def _verify_predict_cl_args(predict_cl_args: Namespace):
    if predict_cl_args.evaluate:
        if len(predict_cl_args.output_configs) == 0:
            raise ValueError(
                "If you want to evaluate, you must specify at least one output config."
                "This is needed to know the target column(s) and values to compute"
                "metrics for (i.e., to compare predicted vs. true values)."
            )


def run_predict(predict_cl_args: Namespace):
    run_folder = get_run_folder_from_model_path(model_path=predict_cl_args.model_path)

    device = maybe_parse_device_from_predict_args(predict_cl_args=predict_cl_args)

    loaded_train_experiment = load_serialized_train_experiment(
        run_folder=run_folder,
        device=device,
    )

    set_log_level_for_eir_loggers(
        log_level=loaded_train_experiment.configs.gc.vl.log_level
    )

    predict_experiment = get_default_predict_experiment(
        loaded_train_experiment=loaded_train_experiment,
        predict_cl_args=predict_cl_args,
        inferred_run_folder=run_folder,
    )

    predict(
        predict_experiment=predict_experiment,
        predict_cl_args=predict_cl_args,
    )
    if predict_experiment.configs.gc.aa.compute_attributions:
        compute_predict_attributions(
            run_folder=run_folder,
            loaded_train_experiment=loaded_train_experiment,
            predict_config=predict_experiment,
        )


def maybe_parse_device_from_predict_args(predict_cl_args: Namespace) -> str:
    device = "cpu"

    if not predict_cl_args.global_configs:
        return device

    global_config_path = predict_cl_args.global_configs[0]
    global_config = load_yaml_config(config_path=global_config_path)

    device_from_config = global_config.get("basic_experiment", {}).get("device", {})
    if device_from_config:
        device = device_from_config

    return device


def predict(
    predict_experiment: "PredictExperiment",
    predict_cl_args: Namespace,
) -> None:
    hook_finalizers = register_pre_evaluation_hooks(
        model=predict_experiment.model,
        global_config=predict_experiment.configs.global_config,
    )

    output_generator = get_prediction_outputs_generator(
        data_loader=predict_experiment.test_dataloader,
        batch_prep_hook=predict_experiment.hooks.step_func_hooks.base_prepare_batch,
        batch_prep_hook_kwargs={"experiment": predict_experiment},
        model=predict_experiment.model,
        with_labels=predict_cl_args.evaluate,
    )

    criteria = train.get_criteria(outputs_as_dict=predict_experiment.outputs)

    loss_func = train.get_loss_callable(
        criteria=criteria,
    )

    predict_results = run_split_evaluation(
        output_generator=output_generator,
        output_objects=predict_experiment.outputs,
        experiment_metrics=predict_experiment.metrics,
        loss_function=loss_func,
        device=predict_experiment.configs.gc.be.device,
        with_labels=predict_cl_args.evaluate,
        missing_ids_per_output=predict_experiment.test_dataset.missing_ids_per_output,
    )

    hook_outputs = deregister_pre_evaluation_hooks(
        hook_finalizers=hook_finalizers,
        evaluation_results=predict_results,
    )

    run_all_eval_hook_analysis(
        hook_outputs=hook_outputs,
        run_folder=Path(predict_cl_args.output_folder),
        iteration="predict",
    )

    if predict_cl_args.evaluate:
        serialize_prediction_metrics(
            output_folder=Path(predict_cl_args.output_folder),
            metrics=predict_results.metrics_with_averages,
        )

        predict_tabular_wrapper_with_labels(
            predict_config=predict_experiment,
            all_predictions=predict_results.gathered_outputs,
            all_labels=predict_results.gathered_labels,
            all_ids=predict_results.ids_per_output,
            predict_cl_args=predict_cl_args,
        )

        predict_survival_wrapper_with_labels(
            predict_config=predict_experiment,
            all_predictions=predict_results.gathered_outputs,
            all_labels=predict_results.gathered_labels,
            all_ids=predict_results.ids_per_output,
            predict_cl_args=predict_cl_args,
        )

    else:
        predict_tabular_wrapper_no_labels(
            predict_config=predict_experiment,
            all_predictions=predict_results.gathered_outputs,
            all_ids=predict_results.ids_per_output,
            predict_cl_args=predict_cl_args,
        )

        predict_survival_wrapper_no_labels(
            predict_config=predict_experiment,
            all_predictions=predict_results.gathered_outputs,
            all_ids=predict_results.ids_per_output,
            predict_cl_args=predict_cl_args,
        )

    predict_sequence_wrapper(
        predict_experiment=predict_experiment,
        output_folder=predict_cl_args.output_folder,
    )

    predict_array_wrapper(
        predict_experiment=predict_experiment,
        output_folder=predict_cl_args.output_folder,
    )


def serialize_prediction_metrics(output_folder: Path, metrics: al_step_metric_dict):
    ensure_path_exists(path=output_folder, is_folder=True)
    with open(str(output_folder / "calculated_metrics.json"), "w") as outfile:
        parsed = _convert_dict_values_to_python_objects(object_=metrics)
        json.dump(obj=parsed, fp=outfile)


def _convert_dict_values_to_python_objects(object_):
    """
    Needed since json does not serialize numpy dtypes natively.
    """

    if isinstance(object_, np.number):
        return object_.item()
    elif isinstance(object_, dict):
        for key, value in object_.items():
            object_[key] = _convert_dict_values_to_python_objects(object_=value)

    return object_


@dataclass
class PredictExperiment:
    configs: Configs
    inputs: al_input_objects_as_dict
    outputs: al_output_objects_as_dict
    predict_specific_cl_args: PredictSpecificCLArgs
    test_dataset: datasets.DiskDataset | datasets.MemoryDataset
    test_dataloader: DataLoader
    model: al_meta_model
    hooks: "PredictHooks"
    metrics: al_metric_record_dict


@dataclass
class PredictHooks:
    step_func_hooks: "PredictStepFunctionHookStages"


@dataclass
class PredictStepFunctionHookStages:
    base_prepare_batch: al_predict_hooks
    model_forward: al_predict_hooks


def get_default_predict_experiment(
    loaded_train_experiment: "LoadedTrainExperiment",
    predict_cl_args: Namespace,
    inferred_run_folder: Path,
) -> PredictExperiment:
    configs_overloaded_for_predict = converge_train_and_predict_configs(
        train_configs=loaded_train_experiment.configs, predict_cl_args=predict_cl_args
    )

    default_train_hooks = loaded_train_experiment.hooks

    if predict_cl_args.evaluate:
        test_ids = gather_all_ids_from_output_configs(
            output_configs=configs_overloaded_for_predict.output_configs
        )

        target_labels = get_target_labels_for_testing(
            configs_overloaded_for_predict=configs_overloaded_for_predict,
            ids=test_ids,
        )
        missing_ids_per_output = target_labels.missing_ids_per_output
    else:
        test_ids = label_setup.gather_all_ids_from_all_inputs(
            input_configs=configs_overloaded_for_predict.input_configs
        )
        target_labels = None
        missing_ids_per_output = MissingTargetsInfo(
            missing_ids_per_modality={},
            all_have_same_set=False,
        )

    test_inputs = set_up_inputs_for_predict(
        test_inputs_configs=configs_overloaded_for_predict.input_configs,
        ids=test_ids,
        hooks=default_train_hooks,
        output_folder=str(inferred_run_folder),
    )

    label_df: pl.DataFrame = (
        target_labels.predict_labels if target_labels else pl.DataFrame()
    )
    outputs_with_predict_paths = _patch_loaded_output_object_configs(
        output_as_dict=loaded_train_experiment.outputs,
        predict_output_configs=configs_overloaded_for_predict.output_configs,
    )

    test_dataset = set_up_default_dataset(
        configs=configs_overloaded_for_predict,
        target_labels_df=label_df,
        inputs_as_dict=test_inputs,
        outputs_as_dict=outputs_with_predict_paths,
        missing_ids_per_output=missing_ids_per_output,
    )
    predict_batch_size = _auto_set_test_batch_size(
        batch_size=configs_overloaded_for_predict.gc.be.batch_size,
        test_set_size=len(test_dataset),
    )

    check_dataset_and_batch_size_compatibility(
        dataset=test_dataset,
        batch_size=predict_batch_size,
        name="Test",
    )
    test_dataloader = DataLoader(
        dataset=test_dataset,
        batch_size=predict_batch_size,
        shuffle=False,
        num_workers=configs_overloaded_for_predict.gc.be.dataloader_workers,
    )

    func = get_meta_model_class_and_kwargs_from_configs
    fusion_model_class, fusion_model_kwargs = func(
        global_config=configs_overloaded_for_predict.global_config,
        fusion_config=configs_overloaded_for_predict.fusion_config,
        inputs_as_dict=test_inputs,
        outputs_as_dict=outputs_with_predict_paths,
        meta_class_getter=get_default_meta_class,
    )

    model = load_model(
        model_path=Path(predict_cl_args.model_path),
        model_class=fusion_model_class,
        model_init_kwargs=fusion_model_kwargs,
        device=configs_overloaded_for_predict.gc.be.device,
        test_mode=True,
        strict_shapes=True,
    )
    assert not model.training

    predict_specific_cl_args = extract_predict_specific_cl_args(
        all_predict_cl_args=predict_cl_args
    )

    default_predict_hooks = _get_default_predict_hooks(train_hooks=default_train_hooks)
    predict_experiment = PredictExperiment(
        configs=configs_overloaded_for_predict,
        inputs=test_inputs,
        outputs=outputs_with_predict_paths,
        predict_specific_cl_args=predict_specific_cl_args,
        test_dataset=test_dataset,
        test_dataloader=test_dataloader,
        model=model,
        hooks=default_predict_hooks,
        metrics=loaded_train_experiment.metrics,
    )

    return predict_experiment


def _patch_loaded_output_object_configs(
    output_as_dict: al_output_objects_as_dict,
    predict_output_configs: Sequence[OutputConfig],
) -> al_output_objects_as_dict:
    """
    This is needed as the loaded output objects have the stripped configs from
    the serializations folder, meaning they have None for e.g. output_source. Here
    we inject the paths as passed in the predict configs.
    """

    outputs_patched = {}

    for output_name, output_object in output_as_dict.items():
        output_object_copy = copy(output_object)
        matching_config = next(
            (
                config
                for config in predict_output_configs
                if config.output_info.output_name == output_name
            ),
            None,
        )
        if matching_config is None:
            raise ValueError(
                f"Could not find output config for output '{output_name}'."
            )

        output_object_copy.output_config = matching_config
        outputs_patched[output_name] = output_object_copy

    return outputs_patched


def _auto_set_test_batch_size(batch_size: int, test_set_size: int) -> int:
    if test_set_size < batch_size:
        logger.warning(
            f"Test set size ({test_set_size}) is smaller than "
            f"batch size ({batch_size}). "
            f"Setting batch size to test set size."
        )
        batch_size = test_set_size
    return batch_size


def _get_default_predict_hooks(train_hooks: "Hooks") -> PredictHooks:
    stages = PredictStepFunctionHookStages(
        base_prepare_batch=[_hook_default_predict_prepare_batch],
        model_forward=train_hooks.step_func_hooks.model_forward,
    )
    predict_hooks = PredictHooks(
        step_func_hooks=stages,
    )

    return predict_hooks


def _hook_default_predict_prepare_batch(
    experiment: "PredictExperiment",
    loader_batch,
    *args,
    **kwargs,
):
    batch = prepare_base_batch_default(
        loader_batch=loader_batch,
        input_objects=experiment.inputs,
        output_objects=experiment.outputs,
        model=experiment.model,
        device=experiment.configs.gc.be.device,
    )

    state_updates = {"batch": batch}

    return state_updates


def extract_predict_specific_cl_args(
    all_predict_cl_args: Namespace,
) -> PredictSpecificCLArgs:
    predict_module_specific_keys = PredictSpecificCLArgs.__dataclass_fields__.keys()

    dataclass_kwargs = {
        k: v
        for k, v in all_predict_cl_args.__dict__.items()
        if k in predict_module_specific_keys
    }
    predict_specific_cl_args_object = PredictSpecificCLArgs(**dataclass_kwargs)

    return predict_specific_cl_args_object


if __name__ == "__main__":
    main()



from argparse import Namespace
from typing import TYPE_CHECKING

from eir.setup.config_validation import validate_input_configs, validate_output_configs

if TYPE_CHECKING:
    from eir.setup.config import Configs


def validate_predict_configs_and_args(
    predict_configs: "Configs", predict_cl_args: Namespace
) -> None:
    validate_input_configs(input_configs=predict_configs.input_configs)
    validate_output_configs(output_configs=predict_configs.output_configs)
    _validate_predict_cl_args_and_config_synergy(
        predict_configs=predict_configs, predict_cl_args=predict_cl_args
    )


def _validate_predict_cl_args_and_config_synergy(
    predict_cl_args: Namespace, predict_configs: "Configs"
) -> None:
    gc = predict_configs.global_config

    if gc.aa.compute_attributions and not predict_cl_args.evaluate:
        raise ValueError(
            "When doing prediction, if compute_attributions is True, "
            "--evaluate in the eirpredict CL arguments must be used. "
            "This is because the currently the true labels are used "
            "to compute the attributions. "
        )



import logging
import random
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    DefaultDict,
    Generator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
import seaborn as sns
import torch
from captum._utils.common import (
    _expand_additional_forward_args,
    _expand_target,
    _format_additional_forward_args,
)
from captum._utils.typing import TargetType
from captum.attr import IntegratedGradients
from captum.attr._utils.approximation_methods import approximation_parameters
from captum.attr._utils.common import _reshape_and_sum
from matplotlib import pyplot as plt
from scipy.stats import bootstrap
from sklearn.preprocessing import KBinsDiscretizer
from torch import Tensor

from eir.setup import schemas
from eir.setup.output_setup_modules.survival_output_setup import (
    ComputedSurvivalOutputInfo,
    SurvivalOutputTypeConfig,
)
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.data_load.label_setup import al_label_transformers_object
    from eir.interpretation.interpretation import SampleAttribution


logger = get_logger(name=__name__)


@dataclass()
class TargetTypeInfo:
    name: str
    type_: str
    cat_is_bce: bool


def get_target_class_name(
    sample_label: torch.Tensor,
    target_transformer: "al_label_transformers_object",
    target_info: TargetTypeInfo,
) -> str:
    """
    Note that we have a special case when working with BCE, where the
    mapping of index -> class name breaks down. In this case, we are always looking
    at index 0 in the produced network outputs (i.e., sample label is patched
    to be always 0) which noes not represent class_ 0 in the transformer,
    but rather score towards the positive class (index 1).
    """

    if target_info.type_ == "con":
        return target_info.name

    tt_it = target_transformer.inverse_transform

    match target_transformer:
        case KBinsDiscretizer():
            bin_idx = sample_label.item()

            bin_edges = target_transformer.bin_edges_[0]

            lower_bound = bin_edges[bin_idx]
            upper_bound = bin_edges[bin_idx + 1]

            return f"time_{lower_bound:.1f}_to_{upper_bound:.1f}"
        case _:
            if not target_info.cat_is_bce:
                cur_trn_label = tt_it([sample_label.item()])[0]
            else:
                cur_trn_label = target_transformer.classes_[1]

    return cur_trn_label


def get_appropriate_target_transformer(
    output_object: Union[ComputedTabularOutputInfo, ComputedSurvivalOutputInfo],
    target_column_name: str,
    target_column_type: str,
) -> "al_label_transformers_object":
    """
    We have this so we can inject the time binning transformer for discrete
    survival outputs. Note that this is coupled with other injections done for the
    interpretation of discrete survival outputs (e.g. injecting the current
    time bin as an event target, which then correctly indexes into a given
    time bin node in the model output).
    """
    target_transformer = output_object.target_transformers[target_column_name]

    match output_object:
        case ComputedSurvivalOutputInfo():
            output_type_info = output_object.output_config.output_type_info
            assert isinstance(output_type_info, SurvivalOutputTypeConfig)

            if target_column_type == "cat":
                target_transformer = output_object.target_transformers[
                    output_type_info.time_column
                ]

    return target_transformer


def stratify_attributions_by_target_classes(
    all_attributions: Sequence["SampleAttribution"],
    target_transformer: "al_label_transformers_object",
    output_name: str,
    target_info: TargetTypeInfo,
) -> DefaultDict[Any, list["SampleAttribution"]]:
    all_attributions_target_class_stratified = defaultdict(list)

    for sample in all_attributions:
        cur_labels_all = sample.sample_info.target_labels
        cur_labels = cur_labels_all[output_name][target_info.name]
        cur_label_name = get_target_class_name(
            sample_label=cur_labels,
            target_transformer=target_transformer,
            target_info=target_info,
        )
        all_attributions_target_class_stratified[cur_label_name].append(sample)

    return all_attributions_target_class_stratified


def plot_attributions_bar(
    df_attributions: pd.DataFrame,
    output_path: Union[str, Path],
    top_n: int = 20,
    title: str = "",
    use_bootstrap: bool = True,
) -> None:
    df_token_top_n = calculate_top_n_tokens(
        df_attributions=df_attributions,
        top_n=top_n,
        use_bootstrap=use_bootstrap,
    )
    df_attributions_top_n_sorted = df_attributions[
        df_attributions["Input"].isin(df_token_top_n.index)
    ]

    order = (
        df_attributions_top_n_sorted.groupby("Input")
        .mean()
        .sort_values(by="Attribution", ascending=False)
        .index
    )

    ax: plt.Axes = sns.barplot(
        data=df_attributions_top_n_sorted,
        x="Attribution",
        y="Input",
        order=order,
        color="#1f77b4",
        capsize=0.1,
        errorbar=("ci", 95),
    )

    plt.tight_layout()
    sns_figure: Optional[plt.Figure | plt.SubFigure] = ax.get_figure()
    assert isinstance(sns_figure, plt.Figure)
    assert sns_figure is not None

    if title:
        ax.set_title(title)

    sns_figure.set_size_inches(10, 0.5 * top_n)
    sns_figure.savefig(fname=output_path, bbox_inches="tight")

    plt.close("all")


def calculate_token_statistics(
    df_attributions: pd.DataFrame,
    use_bootstrap: bool = True,
    n_bootstraps: int = 1000,
) -> pd.DataFrame:
    df_token_stats = pd.DataFrame()
    log_once_callable = get_log_once_callable(logger_=logger)

    for input_feature in df_attributions["Input"].unique():
        feature_attributions = df_attributions[
            df_attributions["Input"] == input_feature
        ]["Attribution"].to_numpy()

        if use_bootstrap:

            if len(np.unique(feature_attributions)) < 2:
                log_once_callable(
                    f"Feature '{input_feature}' has less than 2 unique attribution "
                    f"values; unable to perform bootstrapping. Calculating mean "
                    f"without bootstrapping and setting CI_Size to 0 as a default. "
                    f"There might be other input features with the same issue, this "
                    f"message will only be logged once per attribution calculation."
                )

                df_token_stats.loc[input_feature, "AttributionMean"] = np.mean(
                    feature_attributions
                )
                df_token_stats.loc[input_feature, "CI_Size"] = np.nan
                continue

            res = bootstrap(
                (feature_attributions,),
                np.mean,
                n_resamples=n_bootstraps,
                method="basic",
            )
            mean = np.mean(feature_attributions)
            ci_low, ci_high = res.confidence_interval.low, res.confidence_interval.high
            ci_size = ci_high - ci_low

            df_token_stats.loc[input_feature, "AttributionMean"] = mean
            df_token_stats.loc[input_feature, "CI_Size"] = ci_size
        else:
            mean = feature_attributions.mean()
            df_token_stats.loc[input_feature, "AttributionMean"] = mean
            df_token_stats.loc[input_feature, "CI_Size"] = 0

    return df_token_stats


def get_log_once_callable(logger_: logging.Logger) -> Callable:
    has_logged = False

    def _do_log_once(msg: str):
        nonlocal has_logged
        if not has_logged:
            logger_.info(msg)
            has_logged = True

    return _do_log_once


def calculate_top_n_tokens(
    df_attributions: pd.DataFrame,
    top_n: int,
    use_bootstrap: bool,
) -> pd.DataFrame:
    df_token_stats = calculate_token_statistics(
        df_attributions=df_attributions,
        use_bootstrap=use_bootstrap,
    )

    if use_bootstrap:
        df_token_stats["AttributionScore"] = df_token_stats["AttributionMean"] - (
            0.5 * df_token_stats["CI_Size"]
        )
    else:
        df_token_stats["AttributionScore"] = df_token_stats["AttributionMean"]

    df_token_stats_top_n = df_token_stats.nlargest(top_n, "AttributionScore")
    df_token_stats_top_n_sorted = df_token_stats_top_n.sort_values(
        by="AttributionScore", ascending=False
    )

    return df_token_stats_top_n_sorted


def get_basic_sample_attributions_to_analyse_generator(
    interpretation_config: schemas.BasicInterpretationConfig,
    all_attributions: Sequence["SampleAttribution"],
) -> Generator["SampleAttribution", None, None]:
    strategy = interpretation_config.interpretation_sampling_strategy
    n_samples = interpretation_config.num_samples_to_interpret

    all_attributions = list(all_attributions)

    if strategy == "first_n":
        base = all_attributions[:n_samples]
    elif strategy == "random_sample":
        base = random.sample(all_attributions, n_samples)
    else:
        raise ValueError(f"Unrecognized option for sequence sampling: {strategy}.")

    manual_samples = interpretation_config.manual_samples_to_interpret
    if manual_samples:
        for attribution in all_attributions:
            if attribution.sample_info.ids in manual_samples:
                base.append(attribution)

    for item in base:
        yield item


class MyIntegratedGradients(IntegratedGradients):
    """
    Only change here is casting the input to float32 for MPS devices.
    """

    def _attribute(
        self,
        inputs: Tuple[Tensor, ...],
        baselines: Tuple[Union[Tensor, int, float], ...],
        target: TargetType = None,
        additional_forward_args: Any = None,
        n_steps: int = 50,
        method: str = "gausslegendre",
        step_sizes_and_alphas: Union[None, Tuple[List[float], List[float]]] = None,
    ) -> Tuple[Tensor, ...]:
        if step_sizes_and_alphas is None:
            # retrieve step size and scaling factor for specified
            # approximation method
            step_sizes_func, alphas_func = approximation_parameters(method)
            step_sizes, alphas = step_sizes_func(n_steps), alphas_func(n_steps)
        else:
            step_sizes, alphas = step_sizes_and_alphas

        # scale features and compute gradients. (batch size is abbreviated as bsz)
        # scaled_features' dim -> (bsz * #steps x inputs[0].shape[1:], ...)
        scaled_features_tpl = tuple(
            torch.cat(
                [baseline + alpha * (input - baseline) for alpha in alphas], dim=0
            ).requires_grad_()
            for input, baseline in zip(inputs, baselines)
        )

        additional_forward_args = _format_additional_forward_args(
            additional_forward_args
        )
        # apply number of steps to additional forward args
        # currently, number of steps is applied only to additional forward arguments
        # that are nd-tensors. It is assumed that the first dimension is
        # the number of batches.
        # dim -> (bsz * #steps x additional_forward_args[0].shape[1:], ...)
        input_additional_args = (
            _expand_additional_forward_args(additional_forward_args, n_steps)
            if additional_forward_args is not None
            else None
        )
        expanded_target = _expand_target(target, n_steps)

        # grads: dim -> (bsz * #steps x inputs[0].shape[1:], ...)
        grads = self.gradient_func(
            forward_fn=self.forward_func,
            inputs=scaled_features_tpl,
            target_ind=expanded_target,
            additional_forward_args=input_additional_args,
        )

        # flattening grads so that we can multiply it with step-size
        # calling contiguous to avoid `memory whole` problems
        scaled_grads = [
            grad.contiguous().view(n_steps, -1)
            * torch.tensor(step_sizes, dtype=torch.float32)  # here cast for MPS
            .view(n_steps, 1)
            .to(grad.device)
            for grad in grads
        ]

        # aggregates across all steps for each tensor in the input tuple
        # total_grads has the same dimensionality as inputs
        total_grads = tuple(
            _reshape_and_sum(
                scaled_grad, n_steps, grad.shape[0] // n_steps, grad.shape[1:]
            )
            for (scaled_grad, grad) in zip(scaled_grads, grads)
        )

        # computes attribution for each tensor in input tuple
        # attributions has the same dimensionality as inputs
        if not self.multiplies_by_inputs:
            attributions = total_grads
        else:
            attributions = tuple(
                total_grad * (input - baseline)
                for total_grad, input, baseline in zip(total_grads, inputs, baselines)
            )
        return attributions


def get_long_format_attribution_df(
    parsed_attributions: dict[str, list[float]]
) -> pd.DataFrame:
    if not isinstance(parsed_attributions, dict):
        raise TypeError("Input must be a dictionary")

    if not parsed_attributions:
        return pd.DataFrame(columns=["Input", "Attribution"])

    series_dict: dict[str, pd.Series] = {}
    for k, v in parsed_attributions.items():
        if not isinstance(v, list):
            raise ValueError(f"Value for key '{k}' must be a list")
        if not all(isinstance(x, (int, float, np.float32, np.float64)) for x in v):
            raise ValueError(f"All values in list {v} for key '{k}' must be numbers")
        series_dict[k] = pd.Series(v)

    df: pd.DataFrame = pd.concat(series_dict)
    df = df.reset_index(level=0).reset_index(drop=True)

    df = df.rename(columns={df.columns[0]: "Input", df.columns[1]: "Attribution"})

    return df



import json
from pathlib import Path
from typing import Optional

import numpy as np

from eir.experiment_io.input_object_io_modules.input_io_utils import (
    load_input_config_from_yaml,
)
from eir.experiment_io.io_utils import load_dataclass
from eir.setup.input_setup_modules.setup_array import (
    ArrayNormalizationStats,
    ComputedArrayInputInfo,
    DataDimensions,
    set_up_array_input_object,
)


def load_array_input_object(serialized_input_folder: Path) -> ComputedArrayInputInfo:
    config_path = serialized_input_folder / "input_config.yaml"
    normalization_stats_path = serialized_input_folder / "normalization_stats.json"
    dtype_path = serialized_input_folder / "dtype.json"
    data_dimensions_path = serialized_input_folder / "data_dimensions.json"

    input_config = load_input_config_from_yaml(input_config_path=config_path)

    normalization_stats: Optional[ArrayNormalizationStats] = None
    if normalization_stats_path.exists():
        normalization_stats = load_dataclass(
            cls=ArrayNormalizationStats,
            file_path=normalization_stats_path,
        )

    dtype_str = json.loads(dtype_path.read_text())
    dtype = np.dtype(dtype_str)

    data_dimensions = load_dataclass(
        cls=DataDimensions,
        file_path=data_dimensions_path,
    )

    loaded_object = set_up_array_input_object(
        input_config=input_config,
        normalization_stats=normalization_stats,
        data_dimensions=data_dimensions,
        dtype=dtype,
    )

    return loaded_object



from sklearn.base import BaseEstimator, TransformerMixin


class IdentityTransformer(BaseEstimator, TransformerMixin):
    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X

    def inverse_transform(self, X):
        return X



from typing import TYPE_CHECKING, Callable, Dict

import torch
from torch import nn

from eir.models.layers.mlp_layers import MLPResidualBlock, ResidualMLPConfig
from eir.models.models_utils import (
    calculate_module_dict_outputs,
    create_multi_task_blocks_with_first_adaptor_block,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    pass

al_features = Callable[[Dict[str, torch.Tensor]], torch.Tensor]


logger = get_logger(__name__)


def default_fuse_features(features: Dict[str, torch.Tensor]) -> torch.Tensor:
    feature_flatten = {k: v.flatten(start_dim=1) for k, v in features.items()}
    return torch.cat(tuple(feature_flatten.values()), dim=1)


class MLPResidualFusionModule(nn.Module):
    def __init__(
        self,
        model_config: ResidualMLPConfig,
        fusion_in_dim: int,
        fusion_callable: al_features = default_fuse_features,
        **kwargs,
    ):
        super().__init__()

        self.model_config = model_config
        self.fusion_in_dim = fusion_in_dim
        self.fusion_callable = fusion_callable

        fusion_resblocks_kwargs = {
            "in_features": self.model_config.fc_task_dim,
            "out_features": self.model_config.fc_task_dim,
            "dropout_p": self.model_config.rb_do,
            "stochastic_depth_p": self.model_config.stochastic_depth_p,
            "full_preactivation": False,
        }
        fusion_modules = create_multi_task_blocks_with_first_adaptor_block(
            num_blocks=self.model_config.layers[0],
            branch_names=("fusion",),
            block_constructor=MLPResidualBlock,
            block_constructor_kwargs=fusion_resblocks_kwargs,
            first_layer_kwargs_overload={"in_features": fusion_in_dim},
        )

        self.fusion_modules = fusion_modules

    @property
    def num_out_features(self) -> int:
        return self.model_config.fc_task_dim

    def forward(self, inputs: Dict[str, torch.Tensor]) -> torch.Tensor:
        fused_features = self.fusion_callable(inputs)
        out = calculate_module_dict_outputs(
            input_=fused_features,
            module_dict=self.fusion_modules,
        )

        return out["fusion"]



from argparse import Namespace
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List

import numpy as np
import pandas as pd
import torch

from eir.setup.output_setup_modules.survival_output_setup import (
    ComputedSurvivalOutputInfo,
)
from eir.setup.schemas import SurvivalOutputTypeConfig
from eir.train_utils.evaluation_modules.evaluation_output_survival import (
    calculate_cox_survival_probs,
    plot_cox_individual_curves,
    plot_cox_risk_stratification,
    plot_cox_survival_curves,
    plot_discrete_individual_survival_curves,
    plot_discrete_risk_stratification,
    plot_discrete_survival_curves,
)
from eir.train_utils.metrics import (
    filter_survival_missing_targets,
    general_torch_to_numpy,
)

if TYPE_CHECKING:
    from eir.predict import PredictExperiment


def predict_survival_wrapper_with_labels(
    predict_config: "PredictExperiment",
    all_predictions: Dict[str, Dict[str, torch.Tensor]],
    all_labels: Dict[str, Dict[str, torch.Tensor]],
    all_ids: Dict[str, Dict[str, List[str]]],
    predict_cl_args: Namespace,
) -> None:
    for output_name, output_object in predict_config.outputs.items():
        if not isinstance(output_object, ComputedSurvivalOutputInfo):
            continue

        output_type_info = output_object.output_config.output_type_info
        assert isinstance(output_type_info, SurvivalOutputTypeConfig)

        time_name = output_type_info.time_column
        event_name = output_type_info.event_column
        model_type = (
            "cox" if output_type_info.loss_function == "CoxPHLoss" else "discrete"
        )

        output_folder = Path(predict_cl_args.output_folder, output_name)
        output_folder.mkdir(parents=True, exist_ok=True)

        ids = all_ids[output_name][event_name]
        model_outputs = all_predictions[output_name][event_name]

        if model_type == "discrete":
            times_binned = all_labels[output_name][time_name].cpu().numpy()
            transformers = output_object.target_transformers
            time_kbins_transformer = transformers[time_name]
            it_func = time_kbins_transformer.inverse_transform
            times = it_func(times_binned.reshape(-1, 1)).flatten()
            times = torch.tensor(times)
        else:
            times = all_labels[output_name][time_name]

        events = all_labels[output_name][event_name]

        filtered = filter_survival_missing_targets(
            model_outputs=model_outputs,
            events=events,
            times=times,
            cur_ids=ids,
        )

        model_outputs = filtered.model_outputs
        events_np = general_torch_to_numpy(tensor=filtered.events)
        events_np = events_np.astype(int)

        times_np = general_torch_to_numpy(tensor=filtered.times)
        ids = filtered.ids

        if model_type == "discrete":
            transformers = output_object.target_transformers
            time_kbins_transformer = transformers[time_name]
            time_bins = time_kbins_transformer.bin_edges_[0]
            time_bins_except_last = time_bins[:-1]

            hazards = torch.sigmoid(model_outputs).cpu().numpy()
            survival_probs = np.cumprod(1 - hazards, 1)

            base_data = {"ID": ids, "Predicted_Risk": hazards[:, -1]}

            event_transformer = transformers[event_name]
            events_untransformed = event_transformer.inverse_transform(events_np)
            base_data.update(
                {
                    time_name: times_np,
                    event_name: events_np,
                    f"{event_name} Untransformed": events_untransformed,
                }
            )

            plot_discrete_survival_curves(
                times=times_np,
                events=events_np,
                predicted_probs=survival_probs,
                time_bins=time_bins_except_last,
                output_folder=output_folder,
            )

            plot_discrete_risk_stratification(
                times=times_np,
                events=events_np,
                predicted_probs=survival_probs,
                time_bins=time_bins_except_last,
                output_folder=output_folder,
            )

            surv_prob_cols = {
                f"Surv_Prob_t{i}": survival_probs[:, i]
                for i in range(survival_probs.shape[1])
            }

            df = pd.DataFrame({**base_data, **surv_prob_cols})

            plot_discrete_individual_survival_curves(
                df=df,
                time_bins=time_bins_except_last,
                output_folder=str(output_folder),
                n_samples=5,
            )

        else:
            risk_scores = model_outputs.cpu().numpy()

            baseline_hazard = output_object.baseline_hazard
            unique_times = output_object.baseline_unique_times
            assert baseline_hazard is not None
            assert unique_times is not None

            baseline_survival = np.exp(-np.cumsum(baseline_hazard))

            max_time = unique_times.max()
            time_points = np.linspace(0, max_time, 100)

            survival_probs = calculate_cox_survival_probs(
                risk_scores=risk_scores,
                unique_times=unique_times,
                baseline_survival=baseline_survival,
                time_points=time_points,
            )

            base_data = {"ID": ids, "Risk_Score": risk_scores.squeeze()}
            base_data.update(
                {
                    time_name: times_np,
                    event_name: events_np,
                }
            )

            plot_cox_survival_curves(
                times=times_np,
                events=events_np,
                risk_scores=risk_scores,
                unique_times=unique_times,
                baseline_survival=baseline_survival,
                time_points=time_points,
                output_folder=output_folder,
            )

            plot_cox_risk_stratification(
                times=times_np,
                events=events_np,
                risk_scores=risk_scores,
                unique_times=unique_times,
                baseline_survival=baseline_survival,
                time_points=time_points,
                output_folder=output_folder,
            )

            surv_prob_cols = {
                f"Surv_Prob_t{i}": survival_probs[:, i]
                for i in range(survival_probs.shape[1])
            }

            df = pd.DataFrame({**base_data, **surv_prob_cols})

            plot_cox_individual_curves(
                df=df,
                time_points=time_points,
                output_folder=str(output_folder),
                n_samples=5,
            )

        csv_path = output_folder / "survival_predictions.csv"
        df.to_csv(csv_path, index=False)


def predict_survival_wrapper_no_labels(
    predict_config: "PredictExperiment",
    all_predictions: Dict[str, Dict[str, torch.Tensor]],
    all_ids: Dict[str, Dict[str, List[str]]],
    predict_cl_args: Namespace,
) -> None:
    for output_name, output_object in predict_config.outputs.items():
        if not isinstance(output_object, ComputedSurvivalOutputInfo):
            continue

        output_type_info = output_object.output_config.output_type_info
        assert isinstance(output_type_info, SurvivalOutputTypeConfig)

        time_name = output_type_info.time_column
        event_name = output_type_info.event_column
        model_type = (
            "cox" if output_type_info.loss_function == "CoxPHLoss" else "discrete"
        )

        output_folder = Path(predict_cl_args.output_folder, output_name)
        output_folder.mkdir(parents=True, exist_ok=True)

        ids = all_ids[output_name][event_name]
        model_outputs = all_predictions[output_name][event_name].cpu()

        if model_type == "discrete":
            transformers = output_object.target_transformers
            time_kbins_transformer = transformers[time_name]
            time_bins = time_kbins_transformer.bin_edges_[0]
            time_bins_except_last = time_bins[:-1]

            hazards = torch.sigmoid(model_outputs).cpu().numpy()
            survival_probs = np.cumprod(1 - hazards, 1)

            df = pd.DataFrame(
                {
                    "ID": ids,
                    "Predicted_Risk": hazards[:, -1],
                }
            )

            for i, t in enumerate(time_bins_except_last):
                df[f"Surv_Prob_t{i}"] = survival_probs[:, i]

            plot_discrete_individual_survival_curves(
                df=df,
                time_bins=time_bins_except_last,
                output_folder=str(output_folder),
                n_samples=5,
            )

        else:
            risk_scores = model_outputs.cpu().numpy()

            baseline_hazard = output_object.baseline_hazard
            unique_times = output_object.baseline_unique_times
            assert baseline_hazard is not None
            assert unique_times is not None

            baseline_survival = np.exp(-np.cumsum(baseline_hazard))

            max_time = unique_times.max()
            time_points = np.linspace(0, max_time, 100)

            survival_probs = calculate_cox_survival_probs(
                risk_scores=risk_scores,
                unique_times=unique_times,
                baseline_survival=baseline_survival,
                time_points=time_points,
            )

            base_data = {
                "ID": ids,
                "Risk_Score": risk_scores.squeeze(),
            }

            surv_prob_cols = {
                f"Surv_Prob_t{i}": survival_probs[:, i]
                for i in range(survival_probs.shape[1])
            }

            df = pd.DataFrame({**base_data, **surv_prob_cols})

            plot_cox_individual_curves(
                df=df,
                time_points=time_points,
                output_folder=str(output_folder),
                n_samples=5,
            )

        csv_path = output_folder / "survival_predictions.csv"
        df.to_csv(csv_path, index=False)



import tempfile
from argparse import Namespace
from copy import deepcopy
from pathlib import Path
from typing import Dict, Generator, Iterable, Literal, Protocol, Sequence, Tuple

from eir.predict_modules.predict_config_validation import (
    validate_predict_configs_and_args,
)
from eir.predict_modules.predict_utils import (
    log_and_raise_missing_or_multiple_config_matching_general,
    log_and_raise_missing_or_multiple_tabular_output_matches,
)
from eir.setup import config, schemas
from eir.setup.config import Configs
from eir.setup.config_setup_modules.config_setup_utils import (
    get_yaml_to_dict_iterator,
    object_to_primitives,
    recursive_dict_inject,
)
from eir.setup.schema_modules.output_schemas_tabular import TabularOutputTypeConfig
from eir.utils.logging import get_logger

al_named_dict_configs = Dict[
    Literal["global_configs", "fusion_configs", "input_configs", "output_configs"],
    Iterable[Dict],
]


logger = get_logger(name=__name__)


def converge_train_and_predict_configs(
    train_configs: Configs, predict_cl_args: Namespace
) -> Configs:
    train_configs_copy = deepcopy(train_configs)

    named_dict_iterators = get_named_predict_dict_iterators(
        predict_cl_args=predict_cl_args
    )

    matched_dict_iterator = get_train_predict_matched_config_generator(
        train_configs=train_configs_copy, named_dict_iterators=named_dict_iterators
    )

    predict_configs_overloaded = overload_train_configs_for_predict(
        matched_dict_iterator=matched_dict_iterator
    )
    validate_predict_configs_and_args(
        predict_configs=predict_configs_overloaded, predict_cl_args=predict_cl_args
    )

    return predict_configs_overloaded


def get_named_predict_dict_iterators(
    predict_cl_args: Namespace,
) -> dict[str, tuple[dict, ...]]:
    target_keys = {
        "global_configs",
        "fusion_configs",
        "input_configs",
        "output_configs",
    }
    dict_of_generators = {}
    for key, value in predict_cl_args.__dict__.items():
        if key in target_keys:
            if not value:
                value = ()
            cur_gen = get_yaml_to_dict_iterator(yaml_config_files=value)
            dict_of_generators[key] = tuple(cur_gen)

    return dict_of_generators


def get_train_predict_matched_config_generator(
    train_configs: Configs,
    named_dict_iterators: dict[str, tuple[dict, ...]],
) -> Generator[Tuple[str, Dict, Dict], None, None]:
    train_keys = set(train_configs.__dict__.keys())

    single_configs = {
        "global_configs": "global_config",
        "fusion_configs": "fusion_config",
    }

    sequence_configs = {"input_configs", "output_configs"}

    for predict_argument_name, predict_dict_iter in named_dict_iterators.items():
        name_in_configs_object = single_configs.get(
            predict_argument_name, predict_argument_name
        )
        assert name_in_configs_object in train_keys

        if predict_dict_iter is None:
            predict_dict_iter = []

        # If not a sequence we can yield directly
        if predict_argument_name in single_configs.keys():
            train_config = getattr(train_configs, name_in_configs_object)
            train_config_as_dict = object_to_primitives(obj=train_config)

            predict_config_as_dict = config.combine_dicts(dicts=predict_dict_iter)

            yield (
                name_in_configs_object,
                train_config_as_dict,
                predict_config_as_dict,
            )

        # Otherwise we have to match the respective ones with each other
        elif predict_argument_name in sequence_configs:
            train_config_sequence = getattr(train_configs, name_in_configs_object)

            for cur_config in train_config_sequence:
                is_delayed = _predict_input_config_should_not_exist_yet(
                    predict_argument_name=predict_argument_name,
                    output_configs=train_configs.output_configs,
                    cur_config=cur_config,
                )

                if is_delayed:
                    continue

                matching_func = get_config_sequence_matching_func(
                    name=name_in_configs_object
                )
                predict_dict_match_from_iter = matching_func(
                    train_config=cur_config, predict_dict_iterator=predict_dict_iter
                )

                cur_train_config_as_dict = object_to_primitives(obj=cur_config)

                yield (
                    name_in_configs_object,
                    cur_train_config_as_dict,
                    predict_dict_match_from_iter,
                )


def _predict_input_config_should_not_exist_yet(
    predict_argument_name: str,
    output_configs: Sequence[schemas.OutputConfig],
    cur_config: schemas.InputConfig,
) -> bool:
    """
    There is strict matching functionality for input and output configs for
    e.g. tabular outputs, but for sequence outputs the input is automatically
    generated in the configuration setup, so we don't enforce that the input
    is being passed to the predict module as a yaml config.
    """
    is_delayed = False
    if predict_argument_name == "input_configs":
        matching_output_configs = _find_output_configs_matching_input(
            output_configs=output_configs,
            input_config=cur_config,
        )

        if matching_output_configs:
            assert len(matching_output_configs) == 1
            matching_output_config = matching_output_configs[0]
            if matching_output_config.output_info.output_type == "sequence":
                is_delayed = True

    return is_delayed


def _find_output_configs_matching_input(
    output_configs: Iterable[schemas.OutputConfig], input_config: schemas.InputConfig
) -> list[schemas.OutputConfig]:
    matching_output_configs = [
        i
        for i in output_configs
        if i.output_info.output_name == input_config.input_info.input_name
    ]
    assert len(matching_output_configs) <= 1

    return matching_output_configs


class InputSequenceMatchingFunction(Protocol):
    def __call__(
        self,
        train_config: schemas.InputConfig,
        predict_dict_iterator: Iterable[dict],
    ) -> dict: ...


class OutputSequenceMatchingFunction(Protocol):
    def __call__(
        self,
        train_config: schemas.OutputConfig,
        predict_dict_iterator: Iterable[dict],
    ) -> dict: ...


def get_config_sequence_matching_func(
    name: str,
) -> InputSequenceMatchingFunction | OutputSequenceMatchingFunction:
    assert name in ("input_configs", "output_configs")

    def _input_configs(
        train_config: schemas.InputConfig,
        predict_dict_iterator: Iterable[dict],
    ) -> dict:
        if not isinstance(train_config, schemas.InputConfig):
            raise TypeError("Expected InputConfig for _input_configs")

        matches = []
        predict_names_and_types = []

        train_input_info = train_config.input_info
        train_input_name = train_input_info.input_name
        train_input_type = train_input_info.input_type

        for predict_input_config_dict in predict_dict_iterator:
            predict_feature_extractor_info = predict_input_config_dict["input_info"]
            predict_input_name = predict_feature_extractor_info["input_name"]
            predict_input_type = predict_feature_extractor_info["input_type"]

            predict_names_and_types.append((predict_input_name, predict_input_type))

            cond_1 = predict_input_name == train_input_name
            cond_2 = predict_input_type == train_input_type

            if all((cond_1, cond_2)):
                matches.append(predict_input_config_dict)

        if len(matches) != 1:
            log_and_raise_missing_or_multiple_config_matching_general(
                train_name=train_input_name,
                train_type=train_input_type,
                matches=matches,
                predict_names_and_types=predict_names_and_types,
                name="input",
            )

        assert len(matches) == 1, matches
        return matches[0]

    def _output_configs(
        train_config: schemas.OutputConfig,
        predict_dict_iterator: Iterable[dict],
    ) -> dict:
        if not isinstance(train_config, schemas.OutputConfig):
            raise TypeError("Expected OutputConfig for _output_configs")

        match train_config.output_info.output_type:
            case "tabular":
                match = _check_matching_tabular_output_configs(
                    train_config=train_config,
                    predict_dict_iterator=predict_dict_iterator,
                )
            case "sequence":
                predict_dict_iterator = (
                    _get_maybe_patched_null_sequence_output_source_for_generation(
                        predict_dict_iterator=predict_dict_iterator
                    )
                )
                match = _check_matching_general_output_configs(
                    train_config=train_config,
                    predict_dict_iterator=predict_dict_iterator,
                )

            case "array":
                match = _check_matching_general_output_configs(
                    train_config=train_config,
                    predict_dict_iterator=predict_dict_iterator,
                )

            case "image":
                match = _check_matching_general_output_configs(
                    train_config=train_config,
                    predict_dict_iterator=predict_dict_iterator,
                )

            case "survival":
                match = _check_matching_general_output_configs(
                    train_config=train_config,
                    predict_dict_iterator=predict_dict_iterator,
                )

            case _:
                raise NotImplementedError(
                    f"Output type '{train_config.output_info.output_type}' "
                    f"not implemented."
                )

        return match

    if name == "input_configs":
        return _input_configs
    return _output_configs


def _get_maybe_patched_null_sequence_output_source_for_generation(
    predict_dict_iterator: Iterable[dict],
) -> Iterable[dict]:
    for output_config in predict_dict_iterator:
        if output_config["output_info"]["output_type"] == "sequence":
            if output_config["output_info"]["output_source"] is None:
                output_name = output_config["output_info"]["output_name"]

                logger.info(
                    f"Got 'None' as output source for sequence output: {output_name}. "
                    f"Patching output source to a temporary directory "
                    f"for generation."
                )

                tmp_dir = tempfile.mkdtemp()
                tmp_file = Path(tmp_dir) / f"{output_name}_seed_file.txt"
                tmp_file.touch()

                output_config["output_info"]["output_source"] = str(tmp_dir)

        yield output_config


def _check_matching_general_output_configs(
    train_config: schemas.OutputConfig, predict_dict_iterator: Iterable[dict]
) -> dict:
    train_output_info = train_config.output_info
    train_output_name = train_output_info.output_name
    train_output_type = train_output_info.output_type

    matches = []
    predict_names_and_types = []

    for predict_output_config_dict in predict_dict_iterator:
        predict_output_info = predict_output_config_dict["output_info"]
        predict_output_name = predict_output_info["output_name"]
        predict_output_type = predict_output_info["output_type"]

        predict_names_and_types.append((predict_output_name, predict_output_type))

        cond_1 = predict_output_name == train_output_name
        cond_2 = predict_output_type == train_output_type

        if all((cond_1, cond_2)):
            matches.append(predict_output_config_dict)

    if len(matches) != 1:
        log_and_raise_missing_or_multiple_config_matching_general(
            train_name=train_output_name,
            train_type=train_output_type,
            matches=matches,
            predict_names_and_types=predict_names_and_types,
            name="output",
        )

    assert len(matches) == 1, matches
    return matches[0]


def _check_matching_tabular_output_configs(
    train_config: schemas.OutputConfig,
    predict_dict_iterator: Iterable[dict],
):
    matches = []
    predict_names_and_types = []

    output_name = train_config.output_info.output_name
    output_type = train_config.output_info.output_type

    output_type_info = train_config.output_type_info
    assert isinstance(output_type_info, TabularOutputTypeConfig)

    train_cat_columns = output_type_info.target_cat_columns
    train_con_columns = output_type_info.target_con_columns

    for predict_output_config_dict in predict_dict_iterator:
        predict_output_info = predict_output_config_dict["output_info"]
        predict_output_name = predict_output_info["output_name"]
        predict_output_type = predict_output_info["output_type"]

        cat_cols = predict_output_config_dict.get("output_type_info", {}).get(
            "target_cat_columns", []
        )
        con_cols = predict_output_config_dict.get("output_type_info", {}).get(
            "target_con_columns", []
        )

        predict_names_and_types.append(
            (predict_output_name, predict_output_type, cat_cols, con_cols)
        )

        cond_1 = predict_output_name == output_name
        cond_2 = predict_output_type == output_type

        cond_3 = train_cat_columns == cat_cols
        cond_4 = train_con_columns == con_cols

        if all((cond_1, cond_2, cond_3, cond_4)):
            matches.append(predict_output_config_dict)

    if len(matches) != 1:
        log_and_raise_missing_or_multiple_tabular_output_matches(
            train_name=output_name,
            train_type=output_type,
            train_cat_columns=train_cat_columns,
            train_con_columns=train_con_columns,
            matches=matches,
            predict_names_and_types=predict_names_and_types,
        )

    assert len(matches) == 1, matches
    return matches[0]


def overload_train_configs_for_predict(
    matched_dict_iterator: Generator[Tuple[str, Dict, Dict], None, None],
) -> Configs:
    main_overloaded_kwargs: dict = {}

    for name, train_config_dict, predict_config_dict_to_inject in matched_dict_iterator:
        _maybe_warn_about_output_folder_overload_from_predict(
            name=name,
            predict_config_dict_to_inject=predict_config_dict_to_inject,
            train_config_dict=train_config_dict,
        )

        overloaded_dict = recursive_dict_inject(
            dict_=train_config_dict, dict_to_inject=predict_config_dict_to_inject
        )
        if name in ("global_config", "fusion_config"):
            main_overloaded_kwargs[name] = overloaded_dict
        elif name in ("input_configs", "output_configs"):
            main_overloaded_kwargs.setdefault(name, [])
            main_overloaded_kwargs.get(name, []).append(overloaded_dict)

    global_configs: list[dict] = [main_overloaded_kwargs.get("global_config", {})]
    global_config_overloaded = config.get_global_config(
        global_configs=global_configs,
    )

    if main_overloaded_kwargs.get("input_configs"):
        input_configs_overloaded = config.get_input_configs(
            input_configs=main_overloaded_kwargs.get("input_configs", [])
        )
    else:
        input_configs_overloaded = []

    fusion_config_overloaded = config.load_fusion_configs(
        [main_overloaded_kwargs.get("fusion_config", {})]
    )

    output_configs_overloaded = config.load_output_configs(
        output_configs=main_overloaded_kwargs.get("output_configs", []),
    )

    train_configs_overloaded = config.Configs(
        global_config=global_config_overloaded,
        input_configs=input_configs_overloaded,
        fusion_config=fusion_config_overloaded,
        output_configs=output_configs_overloaded,
    )

    train_configs_overloaded_with_seq_outputs = (
        config.get_configs_object_with_seq_output_configs(
            configs=train_configs_overloaded,
        )
    )

    return train_configs_overloaded_with_seq_outputs


def _maybe_warn_about_output_folder_overload_from_predict(
    name: str, predict_config_dict_to_inject: Dict, train_config_dict: Dict
) -> None:
    if name == "global_config" and "output_folder" in predict_config_dict_to_inject:
        output_folder_from_predict = predict_config_dict_to_inject["output_folder"]
        output_folder_from_train = train_config_dict["output_folder"]

        if output_folder_from_predict == output_folder_from_train:
            return

        logger.warning(
            "output_folder in '%s' will be replaced with '%s' from given prediction "
            "output configuration ('%s'). If this is intentional, you can ignore this "
            "message but most likely this is a mistake and will cause an error. "
            "Resolution: Remove 'output_folder' from "
            "the global output config as it will "
            "be automatically looked up based on the experiment.",
            train_config_dict,
            output_folder_from_predict,
            predict_config_dict_to_inject,
        )



from itertools import chain

import torch
import torch.optim as optim

# Parts of the code are modifications of Pytorch's AdamW optimizer
# Parts of the code are modifications of code from https://github.com/jiaweizzhao/GaLore/blob/master/galore_torch/galore_projector.py


class SOAP(optim.Optimizer):
    """
    Implements SOAP algorithm (https://arxiv.org/abs/2409.11321).

    Parameters:
        params (`Iterable[nn.parameter.Parameter]`):
            Iterable of parameters to optimize or dictionaries defining parameter groups.
        lr (`float`, *optional*, defaults to 0.003):
            The learning rate to use.
        betas (`Tuple[float,float]`, *optional*, defaults to `(0.95, 0.95)`):
            Adam's betas parameters (b1, b2).
        shampoo_beta (`float`, *optional*, defaults to -1):
            If >= 0, use this beta for the preconditioner (L and R in paper, state['GG'] below) moving average instead of betas[1].
        eps (`float`, *optional*, defaults to 1e-08):
            Adam's epsilon for numerical stability.
        weight_decay (`float`, *optional*, defaults to 0.01): weight decay coefficient.
        precondition_frequency (`int`, *optional*, defaults to 10):
            How often to update the preconditioner.
        max_precond_dim (`int`, *optional*, defaults to 10000):
            Maximum dimension of the preconditioner.
            Set to 10000, so that we exclude most common vocab sizes while including layers.
        merge_dims (`bool`, *optional*, defaults to `False`):
            Whether or not to merge dimensions of the preconditioner.
        precondition_1d (`bool`, *optional*, defaults to `False`):
            Whether or not to precondition 1D gradients.
        normalize_grads (`bool`, *optional*, defaults to `False`):
            Whether or not to normalize gradients per layer.
            Helps at large precondition_frequency (~100 in our experiments),
            but hurts performance at small precondition_frequency (~10 in our experiments).
        data_format (`str`, *optional*, defaults to `channels_first`):
            Data format of the input for convolutional layers.
            Should be "channels_last" for data_format of NHWC and "channels_first" for NCHW.
        correct_bias (`bool`, *optional*, defaults to `True`):
            Whether or not to use bias correction in Adam.
    """

    def __init__(
        self,
        params,
        lr: float = 3e-3,
        betas=(0.95, 0.95),
        shampoo_beta: float = -1,
        eps: float = 1e-8,
        weight_decay: float = 0.01,
        precondition_frequency: int = 10,
        max_precond_dim: int = 10000,  #
        merge_dims: bool = False,
        # Merge dimensions till the product of the dimensions is less than or equal to max_precond_dim.
        precondition_1d: bool = False,
        normalize_grads: bool = False,
        data_format: str = "channels_first",
        correct_bias: bool = True,
    ):
        defaults = {
            "lr": lr,
            "betas": betas,
            "shampoo_beta": shampoo_beta,
            "eps": eps,
            "weight_decay": weight_decay,
            "precondition_frequency": precondition_frequency,
            "max_precond_dim": max_precond_dim,
            "merge_dims": merge_dims,
            "precondition_1d": precondition_1d,
            "normalize_grads": normalize_grads,
            "correct_bias": correct_bias,
        }
        super().__init__(params, defaults)
        self._data_format = data_format

    def merge_dims(self, grad, max_precond_dim):
        """
        Merges dimensions of the gradient tensor till the product of the dimensions is less than or equal to max_precond_dim.
        """
        assert self._data_format in ["channels_first", "channels_last"]
        if self._data_format == "channels_last" and grad.dim() == 4:
            grad = grad.permute(0, 3, 1, 2)
        shape = grad.shape
        new_shape = []

        curr_shape = 1
        for sh in shape:
            temp_shape = curr_shape * sh
            if temp_shape > max_precond_dim:
                if curr_shape > 1:
                    new_shape.append(curr_shape)
                    curr_shape = sh
                else:
                    new_shape.append(sh)
                    curr_shape = 1
            else:
                curr_shape = temp_shape

        if curr_shape > 1 or len(new_shape) == 0:
            new_shape.append(curr_shape)

        new_grad = grad.reshape(new_shape)
        return new_grad

    @torch.no_grad()
    def step(self):
        """
        Performs a single optimization step.

        Arguments:
            closure (`Callable`, *optional*): A closure that reevaluates the model and returns the loss.
        """
        loss = None

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad

                state = self.state[p]

                if "step" not in state:
                    state["step"] = 0

                    # State initialization
                if "exp_avg" not in state:
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(grad)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(grad)

                if "Q" not in state:
                    self.init_preconditioner(
                        grad,
                        state,
                        precondition_frequency=group["precondition_frequency"],
                        precondition_1d=group["precondition_1d"],
                        shampoo_beta=(
                            group["shampoo_beta"]
                            if group["shampoo_beta"] >= 0
                            else group["betas"][1]
                        ),
                        max_precond_dim=group["max_precond_dim"],
                        merge_dims=group["merge_dims"],
                    )
                    self.update_preconditioner(
                        grad,
                        state,
                        max_precond_dim=group["max_precond_dim"],
                        merge_dims=group["merge_dims"],
                        precondition_1d=group["precondition_1d"],
                    )
                    continue  # first step is skipped so that we never use the current gradients in the projection.

                # Projecting gradients to the eigenbases of Shampoo's preconditioner
                # i.e. projecting to the eigenbases of matrices in state['GG']
                grad_projected = self.project(
                    grad,
                    state,
                    merge_dims=group["merge_dims"],
                    max_precond_dim=group["max_precond_dim"],
                )

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                # Decay the first and second moment running average coefficient
                # In-place operations to update the averages at the same time
                exp_avg.mul_(beta1).add_(grad, alpha=(1.0 - beta1))
                exp_avg_sq.mul_(beta2).add_(
                    grad_projected.square(), alpha=(1.0 - beta2)
                )

                denom = exp_avg_sq.sqrt().add_(group["eps"])

                # Projecting the exponential moving average of gradients to the eigenbases of Shampoo's preconditioner
                # i.e. projecting to the eigenbases of matrices in state['GG']
                exp_avg_projected = self.project(
                    exp_avg,
                    state,
                    merge_dims=group["merge_dims"],
                    max_precond_dim=group["max_precond_dim"],
                )

                step_size = group["lr"]
                if group["correct_bias"]:
                    bias_correction1 = 1.0 - beta1 ** (state["step"])
                    bias_correction2 = 1.0 - beta2 ** (state["step"])
                    step_size = step_size * (bias_correction2**0.5) / bias_correction1

                # Projecting back the preconditioned (by Adam) exponential moving average of gradients
                # to the original space
                norm_grad = self.project_back(
                    exp_avg_projected / denom,
                    state,
                    merge_dims=group["merge_dims"],
                    max_precond_dim=group["max_precond_dim"],
                )

                if group["normalize_grads"]:
                    norm_grad = norm_grad / (1e-30 + torch.mean(norm_grad**2) ** 0.5)

                p.add_(norm_grad, alpha=-step_size)

                # From AdamW code: Just adding the square of the weights to the loss function is *not*
                # the correct way of using L2 regularization/weight decay with Adam,
                # since that will interact with the m and v parameters in strange ways.
                #
                # Instead we want to decay the weights in a manner that doesn't interact
                # with the m/v parameters. This is equivalent to adding the square
                # of the weights to the loss with plain (non-momentum) SGD.
                # Add weight decay at the end (fixed version)
                if group["weight_decay"] > 0.0:
                    p.add_(p, alpha=(-group["lr"] * group["weight_decay"]))

                # Update is done after the gradient step to avoid using current gradients in the projection.
                self.update_preconditioner(
                    grad,
                    state,
                    max_precond_dim=group["max_precond_dim"],
                    merge_dims=group["merge_dims"],
                    precondition_1d=group["precondition_1d"],
                )

        return loss

    def init_preconditioner(
        self,
        grad,
        state,
        precondition_frequency=10,
        shampoo_beta=0.95,
        max_precond_dim=10000,
        precondition_1d=False,
        merge_dims=False,
    ):
        """
        Initializes the preconditioner matrices (L and R in the paper).
        """
        state["GG"] = (
            []
        )  # Will hold all the preconditioner matrices (L and R in the paper).
        if grad.dim() == 1:
            if not precondition_1d or grad.shape[0] > max_precond_dim:
                state["GG"].append([])
            else:
                state["GG"].append(
                    torch.zeros(grad.shape[0], grad.shape[0], device=grad.device)
                )
        else:
            if merge_dims:
                grad = self.merge_dims(grad, max_precond_dim)

            for sh in grad.shape:
                if sh > max_precond_dim:
                    state["GG"].append([])
                else:
                    state["GG"].append(torch.zeros(sh, sh, device=grad.device))

        state["Q"] = None  # Will hold all the eigenbases of the preconditioner.
        state["precondition_frequency"] = precondition_frequency
        state["shampoo_beta"] = shampoo_beta

    def project(self, grad, state, merge_dims=False, max_precond_dim=10000):
        """
        Projects the gradient to the eigenbases of the preconditioner.
        """
        original_shape = grad.shape
        if merge_dims:
            if grad.dim() == 4 and self._data_format == "channels_last":
                permuted_shape = grad.permute(0, 3, 1, 2).shape
            grad = self.merge_dims(grad, max_precond_dim)

        for mat in state["Q"]:
            if len(mat) > 0:
                grad = torch.tensordot(
                    grad,
                    mat,
                    dims=[[0], [0]],
                )
            else:
                permute_order = list(range(1, len(grad.shape))) + [0]
                grad = grad.permute(permute_order)

        if merge_dims:
            if self._data_format == "channels_last" and len(original_shape) == 4:
                grad = grad.reshape(permuted_shape).permute(0, 2, 3, 1)
            else:
                grad = grad.reshape(original_shape)
        return grad

    def update_preconditioner(
        self,
        grad,
        state,
        max_precond_dim=10000,
        merge_dims=False,
        precondition_1d=False,
    ):
        """
        Updates the preconditioner matrices and the eigenbases (L, R, Q_L, Q_R in the paper).
        """
        if grad.dim() == 1:
            if precondition_1d and grad.shape[0] <= max_precond_dim:
                state["GG"][0].lerp_(
                    grad.unsqueeze(1) @ grad.unsqueeze(0), 1 - state["shampoo_beta"]
                )
        else:
            if merge_dims:
                new_grad = self.merge_dims(grad, max_precond_dim)
                for idx, sh in enumerate(new_grad.shape):
                    if sh <= max_precond_dim:
                        outer_product = torch.tensordot(
                            new_grad,
                            new_grad,
                            dims=[
                                [
                                    *chain(
                                        range(idx), range(idx + 1, len(new_grad.shape))
                                    )
                                ]
                            ]
                            * 2,
                        )
                        state["GG"][idx].lerp_(outer_product, 1 - state["shampoo_beta"])
            else:
                for idx, sh in enumerate(grad.shape):
                    if sh <= max_precond_dim:
                        outer_product = torch.tensordot(
                            grad,
                            grad,
                            # Contracts across all dimensions except for k.
                            dims=[[*chain(range(idx), range(idx + 1, len(grad.shape)))]]
                            * 2,
                        )
                        state["GG"][idx].lerp_(outer_product, 1 - state["shampoo_beta"])

        if state["Q"] is None:
            state["Q"] = self.get_orthogonal_matrix(state["GG"])
        if state["step"] > 0 and state["step"] % state["precondition_frequency"] == 0:
            state["Q"] = self.get_orthogonal_matrix_QR(
                state, max_precond_dim, merge_dims
            )

    def project_back(self, grad, state, merge_dims=False, max_precond_dim=10000):
        """
        Projects the gradient back to the original space.
        """
        original_shape = grad.shape
        if merge_dims:
            if self._data_format == "channels_last" and grad.dim() == 4:
                permuted_shape = grad.permute(0, 3, 1, 2).shape
            grad = self.merge_dims(grad, max_precond_dim)
        for mat in state["Q"]:
            if len(mat) > 0:
                grad = torch.tensordot(
                    grad,
                    mat,
                    dims=[[0], [1]],
                )
            else:
                permute_order = list(range(1, len(grad.shape))) + [0]
                grad = grad.permute(permute_order)

        if merge_dims:
            if self._data_format == "channels_last" and len(original_shape) == 4:
                grad = grad.reshape(permuted_shape).permute(0, 2, 3, 1)
            else:
                grad = grad.reshape(original_shape)
        return grad

    def get_orthogonal_matrix(self, mat):
        """
        Computes the eigenbases of the preconditioner using torch.linalg.eigh decomposition.
        """
        matrix = []
        for m in mat:
            if len(m) == 0:
                matrix.append([])
                continue
            if m.data.dtype != torch.float:
                float_data = False
                original_type = m.data.dtype
                original_device = m.data.device
                matrix.append(m.data.float())
            else:
                float_data = True
                matrix.append(m.data)

        final = []
        for m in matrix:
            if len(m) == 0:
                final.append([])
                continue
            try:
                _, Q = torch.linalg.eigh(
                    m + 1e-30 * torch.eye(m.shape[0], device=m.device)
                )
            except:
                _, Q = torch.linalg.eigh(
                    m.to(torch.float64) + 1e-30 * torch.eye(m.shape[0], device=m.device)
                )
                Q = Q.to(m.dtype)
            Q = torch.flip(Q, [1])

            if not float_data:
                Q = Q.to(original_device).type(original_type)
            final.append(Q)
        return final

    def get_orthogonal_matrix_QR(self, state, max_precond_dim=10000, merge_dims=False):
        """
        Computes the eigenbases of the preconditioner using one round of power iteration
        followed by torch.linalg.qr decomposition.
        """
        precond_list = state["GG"]
        orth_list = state["Q"]

        matrix = []
        orth_matrix = []
        for m, o in zip(precond_list, orth_list):
            if len(m) == 0:
                matrix.append([])
                orth_matrix.append([])
                continue
            if m.data.dtype != torch.float:
                float_data = False
                original_type = m.data.dtype
                original_device = m.data.device
                matrix.append(m.data.float())
                orth_matrix.append(o.data.float())
            else:
                float_data = True
                matrix.append(m.data.float())
                orth_matrix.append(o.data.float())

        orig_shape = state["exp_avg_sq"].shape
        if self._data_format == "channels_last" and len(orig_shape) == 4:
            permuted_shape = state["exp_avg_sq"].permute(0, 3, 1, 2).shape
        if merge_dims:
            exp_avg_sq = self.merge_dims(state["exp_avg_sq"], max_precond_dim)
        else:
            exp_avg_sq = state["exp_avg_sq"]

        final = []
        for ind, (m, o) in enumerate(zip(matrix, orth_matrix)):
            if len(m) == 0:
                final.append([])
                continue
            est_eig = torch.diag(o.T @ m @ o)
            sort_idx = torch.argsort(est_eig, descending=True)
            exp_avg_sq = exp_avg_sq.index_select(ind, sort_idx)
            o = o[:, sort_idx]
            power_iter = m @ o
            Q, _ = torch.linalg.qr(power_iter)

            if not float_data:
                Q = Q.to(original_device).type(original_type)
            final.append(Q)

        if merge_dims:
            if self._data_format == "channels_last" and len(orig_shape) == 4:
                exp_avg_sq = exp_avg_sq.reshape(permuted_shape).permute(0, 2, 3, 1)
            else:
                exp_avg_sq = exp_avg_sq.reshape(orig_shape)

        state["exp_avg_sq"] = exp_avg_sq
        return final



from pathlib import Path
from typing import TYPE_CHECKING, Optional, Protocol

import numpy as np

from eir.data_load.label_setup import al_label_transformers_object
from eir.interpretation.interpretation_utils import (
    TargetTypeInfo,
    get_target_class_name,
)

if TYPE_CHECKING:
    from eir.interpretation.interpretation import SampleAttribution


def analyze_array_input_attributions(
    attribution_outfolder: Path,
    all_attributions: dict[str, np.ndarray],
):
    for target_class, value in all_attributions.items():
        np.save(
            file=str(attribution_outfolder / f"{target_class}.npy"),
            arr=value,
            allow_pickle=True,
        )


class ArrayConsumerCallable(Protocol):
    def __call__(
        self,
        attribution: Optional["SampleAttribution"],
    ) -> Optional[dict[str, np.ndarray]]: ...


def get_array_sum_consumer(
    target_transformer: "al_label_transformers_object",
    input_name: str,
    output_name: str,
    target_info: TargetTypeInfo,
) -> ArrayConsumerCallable:
    results: dict[str, np.ndarray] = {}
    n_samples: dict[str, int] = {}

    target_name = target_info.name

    def _consumer(
        attribution: Optional["SampleAttribution"],
    ) -> Optional[dict[str, np.ndarray]]:
        nonlocal results
        nonlocal n_samples

        if attribution is None:
            for key, value in results.items():
                results[key] = value / n_samples[key]
            return results

        sample_target_labels = attribution.sample_info.target_labels

        cur_label_name = get_target_class_name(
            sample_label=sample_target_labels[output_name][target_name],
            target_transformer=target_transformer,
            target_info=target_info,
        )

        sample_acts = attribution.sample_attributions[input_name].squeeze()
        if cur_label_name not in results:
            results[cur_label_name] = sample_acts
            n_samples[cur_label_name] = 1
        else:
            results[cur_label_name] += sample_acts
            n_samples[cur_label_name] += 1

        return None

    return _consumer



from eir.models.input.omics.omics_models import (
    al_omics_model_configs,
    al_omics_model_types,
    al_omics_models,
    get_omics_model_class,
    get_omics_model_init_kwargs,
)
from eir.setup.input_setup_modules.common import DataDimensions


def get_omics_model_from_model_config(
    model_init_config: al_omics_model_configs,
    data_dimensions: DataDimensions,
    model_type: al_omics_model_types,
) -> al_omics_models:
    omics_model_class = get_omics_model_class(model_type=model_type)
    model_init_kwargs = get_omics_model_init_kwargs(
        model_type=model_type,
        model_config=model_init_config,
        data_dimensions=data_dimensions,
    )
    omics_model = omics_model_class(**model_init_kwargs)  # type: ignore

    if model_type == "cnn":
        size_after_conv = omics_model.data_size_after_conv
        assert isinstance(size_after_conv, int)
        assert size_after_conv >= 1

    return omics_model



import sys
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, Type

import numpy as np
import polars as pl
import torch
from numpy.typing import DTypeLike

from eir.utils.logging import get_logger

logger = get_logger(name=__name__)

NULL_STRINGS = frozenset(("", "None", "null", "nan", "__NULL__"))


class ColumnType(Enum):
    NUMERIC = "numeric"
    STRING = "string"
    PATH = "path"
    FIXED_ARRAY = "fixed_array"
    VAR_ARRAY = "var_array"
    OBJECT = "object"


@dataclass
class ColumnInfo:
    name: str
    dtype: ColumnType
    original_dtype: str
    index: int
    array_shape: Optional[tuple[int, ...]] = None
    inner_dtype: Optional[str] = None
    torch_dtype: Optional[torch.dtype] = None
    np_dtype: Optional[DTypeLike] = None


class HybridStorage:
    def __init__(self):
        self.numeric_float_data: Optional[torch.Tensor] = None
        self.numeric_int_data: Optional[torch.Tensor] = None
        self.string_data: Optional[np.ndarray] = None
        self.path_data: Optional[np.ndarray] = None
        self.fixed_array_data: Optional[list[torch.Tensor]] = None
        self.var_array_data: Optional[list[np.ndarray]] = None
        self.object_data: Optional[dict[str, list[object]]] = None

        self.numeric_float_columns: list[ColumnInfo] = []
        self.numeric_int_columns: list[ColumnInfo] = []
        self.string_columns: list[ColumnInfo] = []
        self.path_columns: list[ColumnInfo] = []
        self.fixed_array_columns: list[ColumnInfo] = []
        self.var_array_columns: list[ColumnInfo] = []
        self.object_columns: list[ColumnInfo] = []
        self._num_rows: int = 0

    def from_polars(self, df: pl.DataFrame) -> None:
        self._num_rows = len(df)

        logger.debug("Converting Polars DataFrame to HybridStorage.")

        for idx, col in enumerate(df.columns):
            series = df.get_column(col)
            dtype = series.dtype
            dtype_str = str(dtype)

            is_fixed, inner_type, shape = _is_fixed_array(dtype=dtype)
            if is_fixed:
                self.fixed_array_columns.append(
                    ColumnInfo(
                        name=col,
                        dtype=ColumnType.FIXED_ARRAY,
                        original_dtype=dtype_str,
                        index=idx,
                        array_shape=shape,
                        inner_dtype=inner_type,
                    )
                )
                continue

            is_var, inner_type = _is_var_array(dtype=dtype)
            if is_var:
                self.var_array_columns.append(
                    ColumnInfo(
                        name=col,
                        dtype=ColumnType.VAR_ARRAY,
                        original_dtype=dtype_str,
                        index=idx,
                        inner_dtype=inner_type,
                    )
                )
                continue

            if _is_numeric_dtype(dtype=dtype):
                is_float = "float" in dtype_str.lower()
                if is_float:
                    torch_dtype = torch.float32
                    np_dtype_float = np.float32
                    self.numeric_float_columns.append(
                        ColumnInfo(
                            name=col,
                            dtype=ColumnType.NUMERIC,
                            original_dtype=dtype_str,
                            index=idx,
                            torch_dtype=torch_dtype,
                            np_dtype=np_dtype_float,
                        )
                    )
                else:
                    torch_dtype = torch.int64
                    np_dtype_int = np.int64
                    self.numeric_int_columns.append(
                        ColumnInfo(
                            name=col,
                            dtype=ColumnType.NUMERIC,
                            original_dtype=dtype_str,
                            index=idx,
                            torch_dtype=torch_dtype,
                            np_dtype=np_dtype_int,
                        )
                    )

            elif _is_path_dtype(name=col, values=series):
                self.path_columns.append(
                    ColumnInfo(
                        name=col,
                        dtype=ColumnType.PATH,
                        original_dtype=dtype_str,
                        index=idx,
                    )
                )

            elif _is_string_dtype(dtype=dtype):
                self.string_columns.append(
                    ColumnInfo(
                        name=col,
                        dtype=ColumnType.STRING,
                        original_dtype=dtype_str,
                        index=idx,
                    )
                )
            else:
                self.object_columns.append(
                    ColumnInfo(
                        name=col,
                        dtype=ColumnType.OBJECT,
                        original_dtype=dtype_str,
                        index=idx,
                    )
                )

        if self.numeric_float_columns:
            float_data = []
            for col_info in self.numeric_float_columns:
                series = df.get_column(name=col_info.name)
                float_data.append(torch.tensor(series.to_numpy(), dtype=torch.float32))

            final_data = None
            if float_data:
                final_data = torch.stack(float_data).to(dtype=torch.float32)
            self.numeric_float_data = final_data

        if self.numeric_int_columns:
            int_data = []
            for col_info in self.numeric_int_columns:
                series = df.get_column(name=col_info.name)
                # will be cast to float64 with NaN
                # copy as the original data is read-only
                numpy_arr = series.to_numpy().copy()

                numpy_arr[np.isnan(numpy_arr)] = -1
                tensor = torch.tensor(numpy_arr, dtype=torch.int64)
                int_data.append(tensor)

            final_data = None
            if int_data:
                final_data = torch.stack(int_data).to(dtype=torch.int64)
            self.numeric_int_data = final_data

        if self.fixed_array_columns:
            fixed_array_data = []
            for col_info in self.fixed_array_columns:
                series = df.get_column(name=col_info.name)
                samples_np = series.to_numpy().astype(col_info.inner_dtype)
                arr = torch.tensor(samples_np)
                fixed_array_data.append(arr)

            self.fixed_array_data = fixed_array_data

        if self.var_array_columns:
            self.var_array_data = []
            for col_info in self.var_array_columns:
                series = df.get_column(name=col_info.name)
                # Store as numpy array for variable length lists
                self.var_array_data.append(series.to_numpy())

        if self.string_columns:
            string_data = []
            for col_info in self.string_columns:
                series = df.get_column(name=col_info.name)
                string_data.append(series.to_numpy())
            self.string_data = np.array(string_data, dtype=np.str_)

        if self.path_columns:
            path_data = []
            for col_info in self.path_columns:
                series = df.get_column(name=col_info.name)
                path_data.append(series.to_numpy())
            self.path_data = np.array(path_data, dtype=np.str_)

        if self.object_columns:
            if self.object_data is None:
                self.object_data = {}

            for col_info in self.object_columns:
                series = df.get_column(name=col_info.name)
                cur_object_data = series.to_list()
                self.object_data[col_info.name] = cur_object_data

    def get_ids(self) -> list[str]:
        id_col_info = next(col for col in self.string_columns if col.name == "ID")
        assert self.string_data is not None
        return self.string_data[self.string_columns.index(id_col_info)].tolist()

    def __len__(self) -> int:
        return self._num_rows

    def __repr__(self) -> str:
        lines = [f"HybridStorage(rows={self._num_rows})"]

        def format_columns(title: str, columns: list[ColumnInfo]) -> list[str]:
            if not columns:
                return []
            result = [f"\n{title} Columns:"]
            for col in columns:
                base = f"  {col.name} ({col.original_dtype})"
                if col.array_shape:
                    base += f" shape={col.array_shape}"
                if col.inner_dtype:
                    base += f" inner_type={col.inner_dtype}"
                result.append(base)
            return result

        sections = [
            ("Numeric Float", self.numeric_float_columns),
            ("Numeric Int", self.numeric_int_columns),
            ("Fixed Array", self.fixed_array_columns),
            ("Variable Array", self.var_array_columns),
            ("String", self.string_columns),
            ("Path", self.path_columns),
            ("Object", self.object_columns),
        ]

        for title, cols in sections:
            lines.extend(format_columns(title, cols))

        return "\n".join(lines)

    def __str__(self) -> str:
        counts = {
            "numeric_float": len(self.numeric_float_columns),
            "numeric_int": len(self.numeric_int_columns),
            "fixed_array": len(self.fixed_array_columns),
            "var_array": len(self.var_array_columns),
            "string": len(self.string_columns),
            "path": len(self.path_columns),
            "object": len(self.object_columns),
        }

        type_counts = [f"{k}={v}" for k, v in counts.items() if v > 0]

        return f"HybridStorage(rows={self._num_rows}, " + ", ".join(type_counts) + ")"

    def check_data(self) -> None:
        """
        Checks for:
            1. Non-empty storage
            2. Valid ID column
            3. At least one non-null value per row
        """
        if self._num_rows == 0:
            raise ValueError(
                "Expected to have at least one sample, but got empty storage. "
                "Possibly there is a mismatch between input IDs and target IDs."
            )

        id_col_info = next(
            (col for col in self.string_columns if col.name == "ID"), None
        )
        if not id_col_info:
            raise ValueError("Could not find required 'ID' column in string columns")

        assert self.string_data is not None
        id_idx = self.string_columns.index(id_col_info)
        id_data = self.string_data[id_idx]

        null_id_mask = np.array([not id_ for id_ in id_data])
        if null_id_mask.any():
            null_indices = np.where(null_id_mask)[0]
            raise ValueError(
                f"Expected all observations to have a sample ID associated "
                f"with them, but got rows with null IDs at indices: "
                f"{null_indices.tolist()}"
            )

        masks = []

        if self.numeric_float_data is not None:
            numeric_mask = torch.isnan(self.numeric_float_data).all(dim=0).numpy()
            masks.append(numeric_mask)

        if self.numeric_int_data is not None:
            numeric_mask = torch.isnan(self.numeric_int_data).all(dim=0).numpy()
            masks.append(numeric_mask)

        if self.string_data is not None:
            string_indices = [i for i in range(len(self.string_columns)) if i != id_idx]
            if string_indices:
                string_mask = np.all(
                    [
                        np.array([not bool(x) for x in self.string_data[i]])
                        for i in string_indices
                    ],
                    axis=0,
                )
                masks.append(string_mask)

        if self.fixed_array_data is not None:
            fixed_array_masks = []
            for col_data in self.fixed_array_data:
                nan_mask = torch.isnan(input=col_data)
                other_dims = tuple(range(1, len(nan_mask.shape)))
                fixed_array_mask = torch.all(input=nan_mask, dim=other_dims).numpy()
                fixed_array_masks.append(fixed_array_mask)

            masks.append(np.all(fixed_array_masks, axis=0))

        if self.var_array_data:
            var_masks: list[np.ndarray] = []
            for arr_data in self.var_array_data:
                cur_mask: list[bool] = []
                for arr in arr_data:
                    cur_mask.append(arr is None or len(arr) == 0)
                var_masks.append(np.array(cur_mask))

            var_array_mask = np.all(var_masks, axis=0)
            masks.append(var_array_mask)

        if self.path_data is not None:
            path_mask = np.all(check_empty_str_arr(data=self.path_data), axis=0)
            masks.append(path_mask)

        if self.object_data:
            object_mask = np.all(
                [
                    np.array([x is None for x in col_data])
                    for col_data in self.object_data.values()
                ],
                axis=0,
            )
            masks.append(object_mask)

        if masks:
            all_null_mask = np.all(masks, axis=0)
            null_rows = np.where(all_null_mask)[0]

            if len(null_rows) > 0:
                bad_id = id_data[null_rows[0]]
                raise ValueError(
                    f"Expected all observations to have at least one input value "
                    f"associated with them, but got empty inputs for ID: {bad_id}"
                )

    def validate_storage(self, name: str = "") -> int:
        self.check_data()

        memory_bytes = get_total_memory(
            numeric_float_data=self.numeric_float_data,
            numeric_int_data=self.numeric_int_data,
            string_data=self.string_data,
            path_data=self.path_data,
            fixed_array_data=self.fixed_array_data,
            var_array_data=self.var_array_data,
            object_data=self.object_data,
        )

        memory_value, memory_unit = format_memory_size(bytes_size=memory_bytes)

        logger.info(
            f"{name} is using {memory_value:.4f}{memory_unit} of memory "
            f"holding {self._num_rows} samples."
        )

        return memory_bytes

    def get_row(self, idx: int) -> dict[str, Any]:
        result: dict[str, Any] = {}

        if self.numeric_float_data is not None:
            row_values = self.numeric_float_data[:, idx]
            row_isnan = torch.isnan(row_values)

            values_np = row_values.cpu().numpy()
            values_np = np.where(row_isnan.cpu().numpy(), np.nan, values_np)

            for i, col_info in enumerate(self.numeric_float_columns):
                # these are already float32, so no special handling needed for NaN
                result[col_info.name] = values_np[i]

        if self.numeric_int_data is not None:
            row_values = self.numeric_int_data[:, idx]
            values_np = row_values.cpu().numpy()

            for i, col_info in enumerate(self.numeric_int_columns):
                value = values_np[i]
                # Convert -1 sentinel back to None for null values
                if value == -1:
                    result[col_info.name] = None
                else:
                    result[col_info.name] = value.item()

        if self.fixed_array_data is not None:
            for i, col_info in enumerate(self.fixed_array_columns):
                result[col_info.name] = self.fixed_array_data[i][idx].numpy()

        if self.var_array_data is not None:
            for i, col_info in enumerate(self.var_array_columns):
                result[col_info.name] = self.var_array_data[i][idx]

        if self.string_data is not None:
            for i, col_info in enumerate(self.string_columns):
                result[col_info.name] = self.string_data[i, idx].item()

        if self.path_data is not None:
            for i, col_info in enumerate(self.path_columns):
                result[col_info.name] = self.path_data[i, idx].item()

        if self.object_data is not None:
            for i, col_info in enumerate(self.object_columns):
                result[col_info.name] = self.object_data[col_info.name][idx]

        return result


def check_empty_str_arr(data: np.ndarray) -> bool:
    str_arr = data.astype(str)
    return (str_arr == "") | (str_arr == "nan")


def _recursive_find_primitive_dtype(dtype: pl.DataType) -> Type[pl.DataType]:
    if hasattr(dtype, "inner"):
        return _recursive_find_primitive_dtype(dtype=dtype.inner)
    return type(dtype)


def _is_numeric_dtype(dtype: pl.DataType) -> bool:
    return dtype.is_numeric()


def _is_fixed_array(
    dtype: pl.DataType,
) -> tuple[bool, Optional[str], Optional[tuple]]:
    if not dtype == pl.Array:
        return False, None, None

    assert isinstance(dtype, pl.Array)

    inner_dtype = dtype.inner
    assert isinstance(inner_dtype, pl.DataType)

    inner_type = _recursive_find_primitive_dtype(dtype=inner_dtype)
    inner_dtype_str = polars_dtype_to_str_dtype(polars_dtype=inner_type)
    shape = dtype.shape

    return True, inner_dtype_str, shape


def _is_var_array(dtype: pl.DataType) -> tuple[bool, Optional[str]]:
    if not dtype == pl.List:
        return False, None

    assert isinstance(dtype, pl.List)
    inner_type = dtype.inner
    return True, str(inner_type)


def _is_path_dtype(name: str, values: pl.Series) -> bool:
    if not isinstance(values[0], str):
        return False
    return any(
        p in str(values[0]).lower()
        for p in ["/var/", "/tmp/", ".csv", ".txt", ".json", ".npy"]
    )


def _is_string_dtype(dtype: pl.DataType) -> bool:
    return dtype == pl.String


def check_two_storages(
    input_storage: HybridStorage,
    target_storage: Optional[HybridStorage],
) -> None:
    if target_storage is None or target_storage._num_rows == 0:
        return

    masks = []

    if target_storage.numeric_float_data is not None:
        numeric_mask = torch.isnan(target_storage.numeric_float_data).all(dim=0).numpy()
        masks.append(numeric_mask)

    if target_storage.numeric_int_data is not None:
        numeric_mask = torch.isnan(target_storage.numeric_int_data).all(dim=0).numpy()
        masks.append(numeric_mask)

    if target_storage.string_data is not None:
        id_idx = next(
            i for i, col in enumerate(target_storage.string_columns) if col.name == "ID"
        )
        string_indices = [
            i for i in range(len(target_storage.string_columns)) if i != id_idx
        ]
        if string_indices:
            string_mask = np.all(
                [
                    np.array([not bool(x) for x in target_storage.string_data[i]])
                    for i in string_indices
                ],
                axis=0,
            )
            masks.append(string_mask)

    if target_storage.fixed_array_data is not None:
        fixed_array_masks = []
        for col_data in target_storage.fixed_array_data:
            nan_mask = torch.isnan(input=col_data)
            other_dims = tuple(range(1, len(nan_mask.shape)))
            fixed_array_mask = torch.all(input=nan_mask, dim=other_dims).numpy()
            fixed_array_masks.append(fixed_array_mask)

        masks.append(np.all(fixed_array_masks, axis=0))

    if target_storage.var_array_data:
        var_masks: list[np.ndarray] = []
        for arr_data in target_storage.var_array_data:
            cur_mask: list[bool] = []
            for arr in arr_data:
                cur_mask.append(arr is None or len(arr) == 0)
            var_masks.append(np.array(cur_mask))

        var_array_mask = np.all(var_masks, axis=0)
        masks.append(var_array_mask)

    if target_storage.object_data:
        object_mask = np.all(
            [
                np.array([x is None for x in col_data])
                for col_data in target_storage.object_data.values()
            ],
            axis=0,
        )
        masks.append(object_mask)

    if masks:
        all_null_mask = np.all(masks, axis=0)
        null_rows = np.where(all_null_mask)[0]
        if len(null_rows) > 0:
            bad_id = target_storage.get_ids()[null_rows[0]]
            raise ValueError(
                f"Expected all observations to have at least one target label "
                f"associated with them, but got empty targets for ID: {bad_id}"
            )

    input_ids = set(input_storage.get_ids())
    target_ids = set(target_storage.get_ids())

    if input_ids != target_ids:
        missing_in_input = target_ids - input_ids
        missing_in_target = input_ids - target_ids
        msg = []
        if missing_in_input:
            msg.append(f"IDs in target but not in input: {missing_in_input}")
        if missing_in_target:
            msg.append(f"IDs in input but not in target: {missing_in_target}")
        raise ValueError(
            "ID mismatch between input and target storage. " + " ".join(msg)
        )


def is_null_value(value: Any) -> bool:
    value_type = type(value)
    if value_type in (float, np.float32, np.float64):
        return np.isnan(value)
    if value_type is str:
        return value in NULL_STRINGS
    if value_type is type(None):
        return True
    if value_type is int:
        return False
    return False


def polars_dtype_to_str_dtype(polars_dtype: Type[pl.DataType]) -> str:
    dtype_map = {
        pl.Float64: "float64",
        pl.Float32: "float32",
        pl.Int64: "int64",
        pl.Int32: "int32",
        pl.Int16: "int16",
        pl.Int8: "int8",
        pl.UInt64: "uint64",
        pl.UInt32: "uint32",
        pl.UInt16: "uint16",
        pl.UInt8: "uint8",
        pl.Boolean: "bool",
        pl.Date: "date",
        pl.Datetime: "datetime",
        pl.Utf8: "str",
    }

    return dtype_map[polars_dtype]


def get_numeric_memory(
    float_data: Optional[torch.Tensor] = None,
    int_data: Optional[torch.Tensor] = None,
) -> int:
    total = 0
    if float_data is not None:
        total += float_data.element_size() * float_data.nelement()
    if int_data is not None:
        total += int_data.element_size() * int_data.nelement()
    return total


def get_fixed_array_memory(data: Optional[list[torch.Tensor]]) -> int:
    if data is None:
        return 0
    return sum(t.element_size() * t.nelement() for t in data)


def get_array_memory(
    data: Optional[np.ndarray],
) -> int:
    memory_usage = 0
    if data is not None:
        memory_usage += data.nbytes
    return memory_usage


def get_path_memory(data: Optional[np.ndarray]) -> int:
    if data is None:
        return 0
    return data.nbytes


def get_var_array_memory(data: Optional[list[np.ndarray]]) -> int:
    if data is None:
        return 0
    return sum(arr.nbytes for arr in data)


def get_object_memory(data: Optional[dict[str, list[object]]]) -> int:
    if data is None:
        return 0
    return sum(sum(sys.getsizeof(obj) for obj in values) for values in data.values())


def format_memory_size(bytes_size: int) -> tuple[float, str]:
    mb_size = bytes_size / (1024**2)
    if mb_size < 1024:
        return mb_size, "MB"
    return mb_size / 1024, "GB"


def get_total_memory(
    numeric_float_data: Optional[torch.Tensor] = None,
    numeric_int_data: Optional[torch.Tensor] = None,
    string_data: Optional[np.ndarray] = None,
    path_data: Optional[np.ndarray] = None,
    fixed_array_data: Optional[list[torch.Tensor]] = None,
    var_array_data: Optional[list[np.ndarray]] = None,
    object_data: Optional[dict[str, list[object]]] = None,
) -> int:
    return sum(
        [
            get_numeric_memory(
                float_data=numeric_float_data,
                int_data=numeric_int_data,
            ),
            get_fixed_array_memory(data=fixed_array_data),
            get_array_memory(data=string_data),
            get_path_memory(data=path_data),
            get_var_array_memory(data=var_array_data),
            get_object_memory(data=object_data),
        ]
    )



__version__ = "0.13.8"

import re
import warnings
from dataclasses import dataclass
from typing import Optional, Protocol, TextIO


class ShowWarningProtocol(Protocol):
    def __call__(
        self,
        message: Warning | str,
        category: type[Warning],
        filename: str,
        lineno: int,
        file: Optional[TextIO] = None,
        line: Optional[str] = None,
    ) -> None: ...


@dataclass
class WarningRule:
    module_pattern: str
    categories: Optional[list[type[Warning]]] = None
    message_pattern: Optional[str] = None


class WarningFilter:
    def __init__(self, rules: list[WarningRule]) -> None:
        self.rules = rules
        self._original_showwarning: ShowWarningProtocol = warnings.showwarning
        self._compiled_rules = [
            (
                re.compile(pattern=rule.module_pattern),
                (
                    re.compile(pattern=rule.message_pattern)
                    if rule.message_pattern
                    else None
                ),
                rule.categories,
            )
            for rule in rules
        ]

    def __call__(
        self,
        message: Warning | str,
        category: type[Warning],
        filename: str,
        lineno: int,
        file: Optional[TextIO] = None,
        line: Optional[str] = None,
    ) -> None:
        for module_pattern, message_pattern, categories in self._compiled_rules:
            if module_pattern.search(filename):
                if categories and category not in categories:
                    continue
                if message_pattern and not message_pattern.search(str(message)):
                    continue
                return

        self._original_showwarning(
            message=message,
            category=category,
            filename=filename,
            lineno=lineno,
            file=file,
            line=line,
        )

    def install(self) -> None:
        warnings.showwarning = self


warning_filter = WarningFilter(
    [
        WarningRule(
            module_pattern=r"ignite/handlers/checkpoint",
            categories=[DeprecationWarning],
        ),
        WarningRule(
            module_pattern=r"torch/distributed/optim",
            message_pattern=r"TorchScript.*support for functional optimizers",
        ),
    ]
)
warning_filter.install()



from copy import deepcopy
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from eir.models.input.sequence.transformer_models import SequenceModelConfig
from eir.setup.input_setup_modules import setup_sequence
from eir.setup.input_setup_modules.torchtext_port.vocab import Vocab
from eir.setup.schemas import InputConfig, OutputConfig, SequenceOutputTypeConfig
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.setup.input_setup import al_input_objects_as_dict
    from eir.setup.output_setup import al_output_objects_as_dict

logger = get_logger(name=__name__, tqdm_compatible=True)


@dataclass()
class ComputedSequenceOutputInfo:
    output_config: OutputConfig
    vocab: Vocab
    embedding_dim: int
    computed_max_length: int
    encode_func: setup_sequence.al_encode_funcs
    tokenizer: Optional[setup_sequence.al_tokenizers]


def set_up_sequence_output(
    output_config: OutputConfig,
    input_objects: "al_input_objects_as_dict",
    *args,
    **kwargs,
) -> ComputedSequenceOutputInfo:
    output_name = output_config.output_info.output_name
    matching_seq_auto_set_up_input_object = input_objects[output_name]

    model_config = matching_seq_auto_set_up_input_object.input_config.model_config
    assert isinstance(model_config, SequenceModelConfig)

    embedding_dim = model_config.embedding_dim
    matching_input_config = matching_seq_auto_set_up_input_object.input_config

    sequence_input_object_func = get_sequence_input_objects_from_output

    vocab, gathered_stats, tokenizer, encode_callable = sequence_input_object_func(
        output_config=output_config, matching_input_config=matching_input_config
    )

    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, SequenceOutputTypeConfig)

    gathered_stats = setup_sequence.possibly_gather_all_stats_from_input(
        prev_gathered_stats=gathered_stats,
        input_source=output_config.output_info.output_source,
        vocab_file=output_type_info.vocab_file,
        split_on=output_type_info.split_on,
        max_length=output_type_info.max_length,
    )

    computed_max_length = setup_sequence.get_max_length(
        max_length_config_value=output_type_info.max_length,
        gathered_stats=gathered_stats,
    )

    sequence_output = ComputedSequenceOutputInfo(
        output_config=output_config,
        vocab=vocab,
        embedding_dim=embedding_dim,
        computed_max_length=computed_max_length,
        encode_func=encode_callable,
        tokenizer=tokenizer,
    )

    return sequence_output


def get_sequence_input_objects_from_output(
    output_config: OutputConfig,
    matching_input_config: InputConfig,
) -> setup_sequence.al_sequence_input_objects_basic:
    gathered_stats = setup_sequence.GatheredSequenceStats()
    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, SequenceOutputTypeConfig)

    tokenizer, gathered_stats = setup_sequence.get_tokenizer(
        input_config=matching_input_config,
        gathered_stats=gathered_stats,
    )

    vocab = setup_sequence.init_vocab(
        source=output_config.output_info.output_source,
        inner_key=output_config.output_info.output_inner_key,
        tokenizer_name=output_type_info.tokenizer,
        split_on=output_type_info.split_on,
        vocab_file=output_type_info.vocab_file,
        min_freq=output_type_info.min_freq,
        gathered_stats=gathered_stats,
        tokenizer=tokenizer,
    )

    encode_func = setup_sequence.get_tokenizer_encode_func(
        tokenizer=tokenizer,
        pytorch_vocab=vocab,
    )

    return vocab, gathered_stats, tokenizer, encode_func


def converge_sequence_input_and_output(
    inputs: "al_input_objects_as_dict",
    outputs: "al_output_objects_as_dict",
) -> "al_input_objects_as_dict":
    inputs_copy = deepcopy(inputs)

    for output_name, output_object in outputs.items():
        cur_out_type_info = output_object.output_config.output_type_info
        if getattr(cur_out_type_info, "sequence_operation", None) not in (
            "autoregressive",
            "mlm",
        ):
            continue

        logger.info(f"Converging input and output for {output_name}.")

        cur_input = inputs_copy[output_name]
        assert isinstance(cur_input, setup_sequence.ComputedSequenceInputInfo)
        assert isinstance(output_object, ComputedSequenceOutputInfo)

        cur_input.computed_max_length = output_object.computed_max_length
        cur_input.vocab = output_object.vocab
        cur_input.tokenizer = output_object.tokenizer
        cur_input.encode_func = output_object.encode_func

        inputs_copy[output_name] = cur_input

    return inputs_copy



from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Sequence, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from aislib.misc_utils import ensure_path_exists
from sklearn.preprocessing import LabelEncoder

from eir.experiment_io.label_transformer_io import load_transformers
from eir.interpretation.interpretation_utils import (
    TargetTypeInfo,
    get_appropriate_target_transformer,
    get_long_format_attribution_df,
    plot_attributions_bar,
    stratify_attributions_by_target_classes,
)
from eir.models.input.tabular.tabular import SimpleTabularModel
from eir.predict_modules.predict_tabular_input_setup import (
    ComputedPredictTabularInputInfo,
)
from eir.setup.input_setup_modules.setup_tabular import ComputedTabularInputInfo
from eir.setup.output_setup_modules.survival_output_setup import (
    ComputedSurvivalOutputInfo,
)
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)
from eir.setup.schemas import TabularInputDataConfig
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.interpretation.interpretation import SampleAttribution
    from eir.predict import PredictExperiment
    from eir.predict_modules.predict_attributions import (
        LoadedTrainExperimentMixedWithPredict,
    )
    from eir.train import Experiment


logger = get_logger(__name__)


def analyze_tabular_input_attributions(
    experiment: Union[
        "Experiment",
        "PredictExperiment",
        "LoadedTrainExperimentMixedWithPredict",
    ],
    input_name: str,
    output_name: str,
    target_info: TargetTypeInfo,
    attribution_outfolder: Path,
    all_attributions: Sequence["SampleAttribution"],
):
    exp = experiment
    input_object = exp.inputs[input_name]
    assert isinstance(
        input_object, (ComputedTabularInputInfo, ComputedPredictTabularInputInfo)
    )

    tabular_type_info_config = input_object.input_config.input_type_info
    assert isinstance(tabular_type_info_config, TabularInputDataConfig)

    cat_columns = tabular_type_info_config.input_cat_columns
    con_columns = tabular_type_info_config.input_con_columns

    tabular_model = experiment.model.input_modules[input_name]
    assert isinstance(tabular_model, SimpleTabularModel)
    attribution_tensor_slices = set_up_tabular_tensor_slices(
        cat_input_columns=cat_columns,
        con_input_columns=con_columns,
        embedding_module=tabular_model,
    )

    parsed_attributions = parse_tabular_attributions_for_feature_importance(
        attribution_tensor_slices=attribution_tensor_slices,
        all_attributions=all_attributions,
        input_name=input_name,
    )
    df_attributions = get_long_format_attribution_df(
        parsed_attributions=parsed_attributions
    )

    plot_attributions_bar(
        df_attributions=df_attributions,
        output_path=attribution_outfolder / "feature_importance.pdf",
        use_bootstrap=False,
    )
    plot_attributions_bar(
        df_attributions=df_attributions,
        output_path=attribution_outfolder
        / "feature_importance_sorted_with_bootstrap.pdf",
        use_bootstrap=True,
    )
    df_attributions.to_csv(attribution_outfolder / "feature_importances.csv")

    output_object = exp.outputs[output_name]
    assert isinstance(
        output_object, (ComputedTabularOutputInfo, ComputedSurvivalOutputInfo)
    )

    target_column_name = target_info.name
    target_column_type = target_info.type_

    target_transformer = get_appropriate_target_transformer(
        output_object=output_object,
        target_column_name=target_column_name,
        target_column_type=target_column_type,
    )

    all_attributions_class_stratified = stratify_attributions_by_target_classes(
        all_attributions=all_attributions,
        target_transformer=target_transformer,
        output_name=output_name,
        target_info=target_info,
    )

    for class_name, class_attributions in all_attributions_class_stratified.items():
        cur_class_outfolder = attribution_outfolder / class_name
        ensure_path_exists(path=cur_class_outfolder, is_folder=True)

        if con_columns:
            cat_to_con_cutoff = get_cat_to_con_cutoff_from_slices(
                slices=attribution_tensor_slices,
                cat_input_columns=cat_columns,
            )
            continuous_attr = _gather_continuous_attributions(
                all_attributions=class_attributions,
                cat_to_con_cutoff=cat_to_con_cutoff,
                input_name=input_name,
            )
            continuous_inputs = _gather_continuous_inputs(
                all_attributions=class_attributions,
                cat_to_con_cutoff=cat_to_con_cutoff,
                con_names=con_columns,
                input_name=input_name,
            )

            plot_tabular_continuous_attribution(
                attributions=continuous_attr,
                df_features=continuous_inputs,
                attribution_output_folder=cur_class_outfolder,
                class_name=class_name,
            )

        cat_act_dfs = []
        for cat_column in cat_columns:
            categorical_attr = _gather_categorical_attributions(
                all_attributions=class_attributions,
                cur_slice=attribution_tensor_slices[cat_column],
                input_name=input_name,
            )
            categorical_inputs = _gather_categorical_inputs(
                all_attributions=class_attributions,
                cat_name=cat_column,
                input_name=input_name,
            )
            cat_column_transformers = load_transformers(
                output_folder=experiment.configs.gc.be.output_folder,
                transformers_to_load={input_name: [cat_column]},
            )

            categorical_inputs_mapped = map_categorical_labels_to_names(
                cat_column_transformers=cat_column_transformers[input_name],
                cat_column=cat_column,
                categorical_inputs=categorical_inputs,
            )

            plot_tabular_categorical_attributions(
                attributions=categorical_attr,
                df_features=categorical_inputs_mapped,
                feature_name_to_plot=cat_column,
                class_name=class_name,
                attribution_output_folder=cur_class_outfolder,
            )

            df_cur_categorical_acts = _parse_categorical_attrs_for_serialization(
                categorical_inputs_mapped=categorical_inputs_mapped,
                attributions_for_input=categorical_attr,
                column_name=cat_column,
            )
            cat_act_dfs.append(df_cur_categorical_acts)

        _save_categorical_acts(
            dfs_categorical_acts_for_class=cat_act_dfs,
            class_name=class_name,
            output_folder=cur_class_outfolder,
        )


def set_up_tabular_tensor_slices(
    cat_input_columns: Sequence[str],
    con_input_columns: Sequence[str],
    embedding_module: SimpleTabularModel,
) -> Dict[str, slice]:
    """
    TODO: Probably better to pass in Dict[column_name, nn.Embedding]
    """
    slices = {}

    current_index = 0
    for cat_column in cat_input_columns:
        cur_embedding = getattr(embedding_module, "embed_" + cat_column)
        cur_embedding_dim = cur_embedding.embedding_dim

        slice_start = current_index
        slice_end = current_index + cur_embedding_dim
        slices[cat_column] = slice(slice_start, slice_end)

        current_index = slice_end

    for con_column in con_input_columns:
        slice_start = current_index
        slice_end = slice_start + 1

        slices[con_column] = slice(slice_start, slice_end)
        current_index = slice_end

    return slices


def get_cat_to_con_cutoff_from_slices(
    slices: Dict[str, slice], cat_input_columns: Sequence[str]
) -> int:
    cutoff = 0

    for cat_column in cat_input_columns:
        cur_cat_slice = slices[cat_column]
        slice_size = cur_cat_slice.stop - cur_cat_slice.start
        cutoff += slice_size

    return cutoff


def parse_tabular_attributions_for_feature_importance(
    attribution_tensor_slices: Dict,
    all_attributions: Sequence["SampleAttribution"],
    input_name: str,
) -> Dict[str, List[float]]:
    """
    Note we need to use abs here to get the absolute feature importance, before
    we sum so different signs don't cancel each other out.
    """
    column_attributions: Dict[str, List[float]] = {
        column: [] for column in attribution_tensor_slices.keys()
    }

    for sample_attribution in all_attributions:
        sample_acts = sample_attribution.sample_attributions[input_name].squeeze(0)

        for column in attribution_tensor_slices.keys():
            cur_slice = attribution_tensor_slices[column]
            cur_column_attributions = np.abs(sample_acts[cur_slice]).sum()
            column_attributions[column].append(cur_column_attributions)

    return column_attributions


def _gather_continuous_inputs(
    all_attributions: Sequence["SampleAttribution"],
    cat_to_con_cutoff: int,
    con_names: Sequence[str],
    input_name: str,
) -> pd.DataFrame:
    con_inputs = []

    for sample in all_attributions:
        cur_full_input = sample.sample_info.inputs[input_name]
        assert len(cur_full_input.shape) == 2

        cur_con_input_part = cur_full_input[:, cat_to_con_cutoff:]
        con_inputs.append(cur_con_input_part)

    con_inputs_array = np.vstack([np.array(i.cpu()) for i in con_inputs])
    df = pd.DataFrame(con_inputs_array, columns=list(con_names))

    return df


def _gather_continuous_attributions(
    all_attributions: Sequence["SampleAttribution"],
    cat_to_con_cutoff: int,
    input_name: str,
) -> np.ndarray:
    con_acts = []

    for sample in all_attributions:
        cur_full_input = sample.sample_attributions[input_name]
        assert len(cur_full_input.shape) == 2
        assert cur_full_input.shape[0] == 1

        cur_con_input_part = cur_full_input[:, cat_to_con_cutoff:]
        con_acts.append(cur_con_input_part)

    array_np = np.vstack(con_acts)

    return array_np


def _gather_categorical_attributions(
    all_attributions: Sequence["SampleAttribution"],
    cur_slice: slice,
    input_name: str,
) -> np.ndarray:
    """
    We need to expand_dims in the case where we only have one input value, otherwise
    shap will not accept a vector.
    """
    cat_acts = []

    for sample in all_attributions:
        cur_full_input = sample.sample_attributions[input_name]
        assert len(cur_full_input.shape) == 2
        assert cur_full_input.shape[0] == 1

        cur_input_squeezed = cur_full_input.squeeze(0)

        cur_cat_input_part = cur_input_squeezed[cur_slice]
        cur_cat_summed_act = np.sum(np.array(cur_cat_input_part))
        cat_acts.append(cur_cat_summed_act)

    array_np = np.vstack(cat_acts)

    if len(array_np.shape) == 1:
        return np.expand_dims(array_np, 1)

    return array_np


def _gather_categorical_inputs(
    all_attributions: Sequence["SampleAttribution"],
    cat_name: str,
    input_name: str,
) -> pd.DataFrame:
    cat_inputs = []

    for sample in all_attributions:
        cur_raw_cat_input = sample.raw_inputs[input_name][cat_name]
        cur_cat_input_part = cur_raw_cat_input
        cat_inputs.append(cur_cat_input_part)

    cat_inputs_array = np.array([np.array(i.cpu()) for i in cat_inputs])
    df = pd.DataFrame(cat_inputs_array, columns=[cat_name])

    return df


def map_categorical_labels_to_names(
    cat_column_transformers: Dict[str, LabelEncoder],
    cat_column: str,
    categorical_inputs: pd.DataFrame,
) -> pd.DataFrame:
    cat_column_transformer = cat_column_transformers[cat_column]

    categorical_inputs_copy = categorical_inputs.copy()

    categorical_inputs_copy[cat_column] = cat_column_transformer.inverse_transform(
        categorical_inputs_copy[cat_column]
    )

    return categorical_inputs_copy


def plot_tabular_categorical_attributions(
    attributions: np.ndarray,
    df_features: pd.DataFrame,
    feature_name_to_plot: str,
    class_name: str,
    attribution_output_folder: Path,
) -> None:
    """
    attribution: (num_samples, num_features)
    features: (num_samples, num_features)
    """

    df_categorical = df_features[[feature_name_to_plot]]

    num_categories = len(df_categorical[feature_name_to_plot].unique())
    fig_width = max(num_categories * 0.5, 8)
    fig_width = min(fig_width, 20)
    fig_height = 6
    fig, ax = plt.subplots(figsize=(fig_width, fig_height))

    category_values = df_categorical[feature_name_to_plot]
    category_attributions = attributions[
        :, df_features.columns.get_loc(feature_name_to_plot)
    ]
    sns.stripplot(
        x=category_values,
        y=category_attributions,
        color="gray",
        alpha=0.5,
        ax=ax,
    )
    sns.pointplot(
        x=category_values,
        y=category_attributions,
        color="black",
        markers=".",
        errorbar=("ci", 95),
        capsize=0.1,
        linestyles="",
        ax=ax,
    )
    ax.set_title(f"Attribution plot for {feature_name_to_plot}")
    ax.set_xlabel(feature_name_to_plot)
    ax.set_ylabel("Attribution (impact on model output)")

    ax.set_xticks(ax.get_xticks())
    ax.set_xticklabels(ax.get_xticklabels(), rotation=90)

    plt.tight_layout()
    name = f"categorical_attributions_{feature_name_to_plot}_{class_name}.pdf"
    plt.savefig(
        str(attribution_output_folder / name),
        bbox_inches="tight",
    )
    plt.close(fig)


def _save_categorical_acts(
    dfs_categorical_acts_for_class: Sequence[pd.DataFrame],
    class_name: str,
    output_folder: Path,
) -> None:
    if len(dfs_categorical_acts_for_class) == 0:
        return None

    df_cat_categorical_acts = pd.concat(dfs_categorical_acts_for_class)
    df_cat_categorical_acts.index.name = "Input_Value"
    csv_name = f"cat_features_{class_name}.csv"
    output_path = str(output_folder / csv_name)
    df_cat_categorical_acts.to_csv(path_or_buf=output_path)


def _parse_categorical_attrs_for_serialization(
    categorical_inputs_mapped: pd.DataFrame,
    attributions_for_input: np.ndarray,
    column_name: str,
) -> pd.DataFrame:
    categorical_inputs_copy = categorical_inputs_mapped.copy()

    assert len(categorical_inputs_copy) == attributions_for_input.shape[0]

    categorical_inputs_copy["Attribution"] = attributions_for_input
    average_effects = categorical_inputs_copy.groupby(by=column_name).mean()
    average_effects["Input_Name"] = average_effects.index.name

    return average_effects


def plot_tabular_continuous_attribution(
    attributions: np.ndarray,
    df_features: pd.DataFrame,
    class_name: str,
    attribution_output_folder: Path,
    top_n_features_to_plot: int = 20,
) -> None:
    """
    attribution: (num_samples, num_features)
    features: (num_samples, num_features)
    """

    feature_importance = np.abs(attributions).mean(axis=0)
    top_n_features = np.argsort(feature_importance)[::-1][:top_n_features_to_plot]

    num_plots = len(top_n_features)
    num_cols = min(num_plots, 4)
    num_rows = (num_plots + num_cols - 1) // num_cols
    fig, axes = plt.subplots(
        nrows=num_rows,
        ncols=num_cols,
        figsize=(4 * num_cols, 4 * num_rows),
        squeeze=False,
    )

    for i, feature_index in enumerate(top_n_features):
        row_index = i // num_cols
        col_index = i % num_cols
        feature_name = df_features.columns[feature_index]

        sns.regplot(
            x=df_features[feature_name],
            y=attributions[:, feature_index],
            color="black",
            scatter_kws={"alpha": 0.7},
            order=1,
            ci=95,
            ax=axes[row_index, col_index],
        )
        axes[row_index, col_index].set_xlabel(feature_name)
        axes[row_index, col_index].set_ylabel("")

    if num_plots < num_rows * num_cols:
        for i in range(num_plots, num_rows * num_cols):
            fig.delaxes(axes[i // num_cols, i % num_cols])

    fig.text(
        -0.01,
        0.5,
        "Attribution (impact on model output)",
        va="center",
        rotation="vertical",
    )

    plt.tight_layout()
    name = f"continuous_attributions_{class_name}.pdf"
    plt.savefig(
        str(attribution_output_folder / name),
        bbox_inches="tight",
    )
    plt.close(fig)



from dataclasses import dataclass, field
from typing import Literal, Sequence, Union

al_cat_loss_names = Literal["CrossEntropyLoss", "BCEWithLogitsLoss"]
al_con_loss_names = Literal[
    "MSELoss",
    "L1Loss",
    "SmoothL1Loss",
    "PoissonNLLLoss",
    "HuberLoss",
]


@dataclass
class TabularOutputTypeConfig:
    """
    :param target_cat_columns:
        Which columns from ``label_file`` to use as categorical targets.

    :param target_con_columns:
        Which columns from ``label_file`` to use as continuous targets.

    :param label_parsing_chunk_size:
        Number of rows to process at time when loading in the ``input_source``. Useful
        when RAM is limited.

    :param cat_label_smoothing:
        Label smoothing to apply to categorical targets.

    :param cat_loss_name:
        Loss function to use for categorical targets. If using `BCEWithLogitsLoss`,
        the targets should all be binary.

    :param con_loss_name:
        Loss function to use for continuous targets.

    :param uncertainty_weighted_mt_loss:
        Whether to use uncertainty weighted loss for multitask / multilabel learning.
    """

    target_cat_columns: Sequence[str] = field(default_factory=list)
    target_con_columns: Sequence[str] = field(default_factory=list)
    label_parsing_chunk_size: Union[None, int] = None
    cat_label_smoothing: float = 0.0
    cat_loss_name: al_cat_loss_names = "CrossEntropyLoss"
    con_loss_name: al_con_loss_names = "MSELoss"
    uncertainty_weighted_mt_loss: bool = False



import reprlib
from collections import defaultdict
from functools import partial
from inspect import signature
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    MutableMapping,
    Optional,
    Type,
)

import torch
from adabelief_pytorch import AdaBelief
from aislib.pytorch_modules import AdaHessian
from torch import nn
from torch.optim import SGD, Adam
from torch.optim.adamw import AdamW
from torch.optim.optimizer import Optimizer
from torch_optimizer import get as get_custom_opt

from eir.models.model_training_utils import add_wd_to_model_params
from eir.setup.setup_utils import get_base_optimizer_names
from eir.train_utils.optimization_modules.soap import SOAP
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.models.model_setup_modules.meta_setup import al_meta_model
    from eir.setup.schemas import GlobalConfig

logger = get_logger(name=__name__, tqdm_compatible=True)


def get_optimizer(
    model: nn.Module,
    loss_callable: Callable,
    global_config: "GlobalConfig",
    extra_modules: Optional[dict[str, nn.Module]] = None,
) -> Optimizer:
    all_params = _get_all_params_to_optimize(
        model=model,
        weight_decay=global_config.opt.wd,
        loss_callable=loss_callable,
        extra_modules=extra_modules,
    )

    optimizer_class = _get_optimizer_class(optimizer_name=global_config.opt.optimizer)
    optimizer_args = _get_constructor_arguments(
        params=all_params,
        global_config=global_config,
        optimizer_class=optimizer_class,
    )
    optimizer = optimizer_class(**optimizer_args)

    logger.debug(
        "Optimizer %s created with arguments %s (note weight decay not included as it "
        "is added manually to parameter groups). Parameter groups %s.",
        optimizer_class,
        get_optimizer_defaults(optimizer=optimizer),
        reprlib.repr(optimizer.param_groups),
    )

    return optimizer


def _get_all_params_to_optimize(
    model: nn.Module,
    weight_decay: float,
    loss_callable: Callable,
    extra_modules: Optional[dict[str, nn.Module]] = None,
) -> list:
    model_params = add_wd_to_model_params(model=model, wd=weight_decay)

    loss_params = []
    if isinstance(loss_callable, nn.Module):
        loss_params = [{"params": p} for p in loss_callable.parameters()]

    extra_params = []
    if extra_modules is not None:
        for module in extra_modules.values():
            extra_params = [{"params": p} for p in module.parameters()]

    return model_params + loss_params + extra_params


def _get_external_optimizers(optimizer_name: str) -> Type[Optimizer]:
    custom_opt = get_custom_opt(name=optimizer_name)
    return custom_opt


def _get_optimizer_class(optimizer_name: str) -> Type[Optimizer]:
    optimizer_getter = _create_optimizer_class_getter(optimizer_name=optimizer_name)
    optimizer_class = optimizer_getter[optimizer_name]
    return optimizer_class


def _create_optimizer_class_getter(
    optimizer_name: str,
) -> MutableMapping[str, Type[Optimizer]]:
    """
    We use an interface with the external optimizer library as a default factory,
    meaning that if we cannot find an optimizer in our base optimizer, we check if
    they exist in the external library.
    """
    default_factory = partial(_get_external_optimizers, optimizer_name)
    optimizer_getter: MutableMapping[str, Type[Optimizer]] = defaultdict(
        default_factory
    )

    base_optimizers = get_base_optimizers_dict()
    optimizer_getter.update(base_optimizers)

    return optimizer_getter


def get_base_optimizers_dict() -> Dict[str, Type[Optimizer]]:
    base_optimizers = {
        "sgdm": SGD,
        "adam": partial(Adam, fused=True),
        "adamw": partial(AdamW, fused=True),
        "adahessian": AdaHessian,
        "adabelief": partial(AdaBelief, print_change_log=False),
        "adabeliefw": partial(AdaBelief, weight_decouple=True, print_change_log=False),
        "soap": SOAP,
    }
    assert set(base_optimizers) == get_base_optimizer_names()
    return base_optimizers


def _get_constructor_arguments(
    params: List, global_config: "GlobalConfig", optimizer_class: Type[Optimizer]
) -> dict[str, Any]:
    base = {
        "params": params,
        "lr": global_config.opt.lr,
        "weight_decay": global_config.opt.wd,
    }

    all_extras = {
        "betas": (global_config.opt.b1, global_config.opt.b2),
        "momentum": 0.9,
        "amsgrad": False,
    }
    accepted_args = signature(optimizer_class).parameters.keys()
    common_extras = {k: v for k, v in all_extras.items() if k in accepted_args}
    return {**base, **common_extras}


def get_optimizer_backward_kwargs(optimizer_name: str) -> dict[str, Any]:
    if optimizer_name == "adahessian":
        return {"create_graph": True}
    return {}


def get_optimizer_defaults(optimizer: Optimizer) -> dict:
    return optimizer.defaults


class AttrDelegatedSWAWrapper(torch.optim.swa_utils.AveragedModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError:
            return getattr(self.module, name)


def maybe_wrap_model_with_swa(
    n_iter_before_swa: Optional[int],
    model: "al_meta_model",
    device: torch.device,
) -> "al_meta_model":
    if n_iter_before_swa is None:
        return model

    swa_model = AttrDelegatedSWAWrapper(model=model, device=device)
    return swa_model



from dataclasses import dataclass
from typing import Dict

from sklearn.preprocessing import StandardScaler

from eir.data_load.label_setup import (
    al_label_transformers,
    al_target_columns,
    merge_target_columns,
)
from eir.setup import schemas
from eir.setup.schema_modules.output_schemas_tabular import (
    TabularOutputTypeConfig,
    al_cat_loss_names,
)
from eir.utils.logging import get_logger

logger = get_logger(name=__name__)

al_num_outputs_per_target = dict[str, int]


@dataclass
class ComputedTabularOutputInfo:
    output_config: schemas.OutputConfig
    num_outputs_per_target: al_num_outputs_per_target
    target_columns: al_target_columns
    target_transformers: al_label_transformers


def set_up_tabular_output(
    output_config: schemas.OutputConfig,
    target_transformers: Dict[str, al_label_transformers],
    *args,
    **kwargs,
) -> ComputedTabularOutputInfo:

    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, TabularOutputTypeConfig)

    cur_target_transformers = target_transformers[output_config.output_info.output_name]
    num_outputs_per_target = set_up_num_outputs_per_target(
        target_transformers=cur_target_transformers,
        cat_loss=output_type_info.cat_loss_name,
    )

    target_columns = merge_target_columns(
        target_con_columns=list(output_type_info.target_con_columns),
        target_cat_columns=list(output_type_info.target_cat_columns),
    )

    tabular_output_info = ComputedTabularOutputInfo(
        output_config=output_config,
        num_outputs_per_target=num_outputs_per_target,
        target_columns=target_columns,
        target_transformers=cur_target_transformers,
    )

    return tabular_output_info


def set_up_num_outputs_per_target(
    target_transformers: al_label_transformers,
    cat_loss: al_cat_loss_names,
) -> al_num_outputs_per_target:
    num_outputs_per_target_dict = {}
    for target_column, transformer in target_transformers.items():
        if isinstance(transformer, StandardScaler):
            num_outputs = 1
        else:
            num_outputs = len(transformer.classes_)

            if num_outputs < 2:
                logger.warning(
                    f"Only {num_outputs} unique values found in categorical label "
                    f"column {target_column} (returned by {transformer}). This means "
                    f"that most likely an error will be raised if e.g. using "
                    f"nn.CrossEntropyLoss as it expects an output dimension of >=2."
                )

            if cat_loss == "BCEWithLogitsLoss":
                if num_outputs != 2:
                    raise ValueError(
                        f"Binary classification loss BCEWithLogitsLoss requires "
                        f"exactly 2 unique values in target column {target_column} "
                        f"but found {num_outputs}."
                    )
                num_outputs = 1

        num_outputs_per_target_dict[target_column] = num_outputs

    return num_outputs_per_target_dict



from pathlib import Path
from typing import TYPE_CHECKING, Callable, Dict, Type, cast

import torch
from torch import nn
from torch.nn import functional as F

from eir.experiment_io.input_object_io_modules.sequence_input_io import (
    load_sequence_input_object,
)
from eir.models.input.sequence.transformer_models import TransformerWrapperModel
from eir.models.model_setup_modules.meta_setup import (
    MetaClassGetterCallable,
    al_meta_model,
    get_default_meta_class,
    get_meta_model_class_and_kwargs_from_configs,
)
from eir.models.model_setup_modules.model_io import load_model, strip_orig_mod_prefix
from eir.models.model_setup_modules.pretrained_setup import (
    overload_meta_model_feature_extractors_with_pretrained,
)
from eir.setup import schemas
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.train_utils.distributed import maybe_make_model_distributed
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.setup.output_setup import al_output_objects_as_dict

al_model_registry = Dict[str, Callable[[str], Type[nn.Module]]]

logger = get_logger(name=__name__)


def get_model(
    global_config: schemas.GlobalConfig,
    inputs_as_dict: al_input_objects_as_dict,
    fusion_config: schemas.FusionConfig,
    outputs_as_dict: "al_output_objects_as_dict",
    meta_class_getter: MetaClassGetterCallable = get_default_meta_class,
) -> al_meta_model:
    meta_class, meta_kwargs = get_meta_model_class_and_kwargs_from_configs(
        global_config=global_config,
        fusion_config=fusion_config,
        inputs_as_dict=inputs_as_dict,
        outputs_as_dict=outputs_as_dict,
        meta_class_getter=meta_class_getter,
    )

    if global_config.m.pretrained_checkpoint:
        logger.info(
            "Loading pretrained checkpoint from '%s'.",
            global_config.m.pretrained_checkpoint,
        )
        loaded_meta_model = load_model(
            model_path=Path(global_config.m.pretrained_checkpoint),
            model_class=meta_class,
            model_init_kwargs=meta_kwargs,
            device=global_config.be.device,
            test_mode=False,
            strict_shapes=global_config.m.strict_pretrained_loading,
        )

        loaded_meta_model = overload_embeddings_with_pretrained(
            model=loaded_meta_model,
            inputs=inputs_as_dict,
            device=global_config.be.device,
            pretrained_checkpoint=global_config.m.pretrained_checkpoint,
        )

        return loaded_meta_model

    input_modules = overload_meta_model_feature_extractors_with_pretrained(
        input_modules=meta_kwargs["input_modules"],
        inputs_as_dict=inputs_as_dict,
        outputs_as_dict=outputs_as_dict,
        meta_class_getter=meta_class_getter,
    )
    meta_kwargs["input_modules"] = input_modules

    meta_model = meta_class(**meta_kwargs)
    device = torch.device(device=global_config.be.device)
    meta_model = meta_model.to(device=device)

    meta_model = maybe_make_model_distributed(
        device=global_config.be.device,
        model=meta_model,
    )

    compiled_model: al_meta_model
    if global_config.m.compile_model:
        compiled_model = cast(al_meta_model, torch.compile(model=meta_model))
    else:
        compiled_model = meta_model

    return compiled_model


def overload_embeddings_with_pretrained(
    model: "al_meta_model",
    inputs: "al_input_objects_as_dict",
    device: str,
    pretrained_checkpoint: str,
) -> "al_meta_model":
    """
    Vocab: From serialized input object
    Embeddings: From loaded model

    - If we have a pretrained checkpoint in global config, we have to initialize inputs
    from that experiment in order to get the vocab per input.

    - If we are using selected ones, we can just use the vocab from the input object
    here directly.

    In both cases, we have to load the pretrained model to grab the embeddings.
    Probably it's enough to just use the torch.load() function here, since it's just
    giving us a dictionary.

    First, let's just assume the global case.
    """

    if not pretrained_checkpoint:
        return model

    any_sequence_inputs = any(
        input_object.input_config.input_info.input_type == "sequence"
        for input_object in inputs.values()
    )
    if not any_sequence_inputs:
        return model

    logger.info(
        f"Overloading embeddings with pretrained checkpoint {pretrained_checkpoint}."
    )

    run_folder = Path(pretrained_checkpoint).parent.parent
    serialized_inputs_folder = (
        run_folder / "serializations/sequence_input_serializations"
    )

    any_serialized_sequence_inputs = serialized_inputs_folder.exists()
    if not any_serialized_sequence_inputs:
        return model

    input_objects_loaded = {}
    for serialized_input_folder in serialized_inputs_folder.iterdir():
        input_name = serialized_input_folder.stem
        serialized_input = load_sequence_input_object(
            serialized_input_folder=serialized_input_folder
        )
        input_objects_loaded[input_name] = serialized_input

    loaded_state_dict = torch.load(
        f=pretrained_checkpoint,
        map_location=device,
        weights_only=True,
    )
    loaded_state_dict = strip_orig_mod_prefix(state_dict=loaded_state_dict)

    for input_name, input_object in inputs.items():
        input_type = input_object.input_config.input_info.input_type
        if input_type != "sequence":
            continue

        assert isinstance(input_object, ComputedSequenceInputInfo)
        cur_vocab = input_object.vocab.get_stoi()
        prev_input_object_vocab = input_objects_loaded[input_name].vocab.get_stoi()

        prev_emb_key = f"input_modules.{input_name}.embedding.weight"
        prev_embeddings = loaded_state_dict[prev_emb_key]

        cur_input_module = model.input_modules[input_name]
        assert isinstance(cur_input_module, TransformerWrapperModel)

        cur_embedding = cur_input_module.embedding.weight
        cur_embedding_copy = cur_embedding.clone().detach()

        for token, idx in cur_vocab.items():
            if token not in prev_input_object_vocab:
                continue

            prev_idx = prev_input_object_vocab[token]
            prev_emb = prev_embeddings[prev_idx]

            cur_emb = cur_embedding_copy[idx]

            if prev_emb.shape != cur_emb.shape:
                logger.warning(
                    f"Shape mismatch for token {token} in input {input_name}."
                    f"Applying average pooling to match dimensions."
                )
                prev_emb = prev_emb.view(1, 1, -1)
                prev_emb = F.adaptive_avg_pool1d(
                    input=prev_emb,
                    output_size=cur_emb.shape[0],
                )
                prev_emb = prev_emb.view(-1)

            cur_embedding_copy[idx] = prev_emb

        cur_input_module.embedding.weight = nn.Parameter(
            data=cur_embedding_copy,
            requires_grad=True,
        )
        logger.info(f"Overloaded embeddings for {input_name}.")

    return model
